import pytest

from dagster import (
    DagsterInvalidDefinitionError,
    DependencyDefinition,
    InputDefinition,
    Int,
    Output,
    OutputDefinition,
    PipelineDefinition,
    composite_solid,
    execute_pipeline,
    lambda_solid,
    pipeline,
    solid,
    usable_as_dagster_type,
)


def builder(graph):
    return graph.add_one(graph.return_one())


@lambda_solid
def return_one():
    return 1


@lambda_solid
def return_two():
    return 2


@lambda_solid
def return_three():
    return 3


@lambda_solid(input_defs=[InputDefinition('num')])
def add_one(num):
    return num + 1


def test_basic_use_case():
    pipeline_def = PipelineDefinition(
        name='basic',
        solid_defs=[return_one, add_one],
        dependencies={'add_one': {'num': DependencyDefinition('return_one')}},
    )

    assert execute_pipeline(pipeline_def).result_for_solid('add_one').output_value() == 2


def test_basic_use_case_with_dsl():
    @pipeline
    def test():
        return add_one(num=return_one())

    assert execute_pipeline(test).result_for_solid('add_one').output_value() == 2


def test_two_inputs_without_dsl():
    @lambda_solid(input_defs=[InputDefinition('num_one'), InputDefinition('num_two')])
    def subtract(num_one, num_two):
        return num_one - num_two

    pipeline_def = PipelineDefinition(
        solid_defs=[subtract, return_two, return_three],
        dependencies={
            'subtract': {
                'num_one': DependencyDefinition('return_two'),
                'num_two': DependencyDefinition('return_three'),
            }
        },
    )

    assert execute_pipeline(pipeline_def).result_for_solid('subtract').output_value() == -1


def test_two_inputs_with_dsl():
    @lambda_solid(input_defs=[InputDefinition('num_one'), InputDefinition('num_two')])
    def subtract(num_one, num_two):
        return num_one - num_two

    @pipeline
    def test():
        return subtract(num_one=return_two(), num_two=return_three())

    assert execute_pipeline(test).result_for_solid('subtract').output_value() == -1


def test_basic_aliasing_with_dsl():
    @pipeline
    def test():
        return add_one.alias('renamed')(num=return_one())

    assert execute_pipeline(test).result_for_solid('renamed').output_value() == 2


def test_diamond_graph():
    @solid(output_defs=[OutputDefinition(name='value_one'), OutputDefinition(name='value_two')])
    def emit_values(_context):
        yield Output(1, 'value_one')
        yield Output(2, 'value_two')

    @lambda_solid(input_defs=[InputDefinition('num_one'), InputDefinition('num_two')])
    def subtract(num_one, num_two):
        return num_one - num_two

    @pipeline
    def diamond_pipeline():
        value_one, value_two = emit_values()
        return subtract(
            num_one=add_one(num=value_one), num_two=add_one.alias('renamed')(num=value_two)
        )

    result = execute_pipeline(diamond_pipeline)

    assert result.result_for_solid('subtract').output_value() == -1


def test_two_cliques():
    @pipeline
    def diamond_pipeline():
        return (return_one(), return_two())

    result = execute_pipeline(diamond_pipeline)

    assert result.result_for_solid('return_one').output_value() == 1
    assert result.result_for_solid('return_two').output_value() == 2


def test_deep_graph():
    @solid(config=Int)
    def download_num(context):
        return context.solid_config

    @lambda_solid(input_defs=[InputDefinition('num')])
    def unzip_num(num):
        return num

    @lambda_solid(input_defs=[InputDefinition('num')])
    def ingest_num(num):
        return num

    @lambda_solid(input_defs=[InputDefinition('num')])
    def subsample_num(num):
        return num

    @lambda_solid(input_defs=[InputDefinition('num')])
    def canonicalize_num(num):
        return num

    @lambda_solid(input_defs=[InputDefinition('num')])
    def load_num(num):
        return num + 3

    @pipeline
    def test():
        return load_num(
            num=canonicalize_num(
                num=subsample_num(num=ingest_num(num=unzip_num(num=download_num())))
            )
        )

    result = execute_pipeline(test, {'solids': {'download_num': {'config': 123}}})
    assert result.result_for_solid('canonicalize_num').output_value() == 123
    assert result.result_for_solid('load_num').output_value() == 126


def test_unconfigurable_inputs_pipeline():
    @usable_as_dagster_type
    class NewType(object):
        pass

    @lambda_solid(input_defs=[InputDefinition('_', NewType)])
    def noop(_):
        pass

    with pytest.raises(DagsterInvalidDefinitionError):
        # NewType is not connect and can not be provided with config
        @pipeline
        def _bad_inputs():
            noop()


def test_dupe_defs_fail():
    @lambda_solid(name='same')
    def noop():
        pass

    @lambda_solid(name='same')
    def noop2():
        pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @pipeline
        def _dupes():
            noop()
            noop2()

    with pytest.raises(DagsterInvalidDefinitionError):
        PipelineDefinition(name='dupes', solid_defs=[noop, noop2])


def test_composite_dupe_defs_fail():
    @lambda_solid
    def noop():
        pass

    @composite_solid(name='same')
    def composite_noop():
        noop()
        noop()
        noop()

    @composite_solid(name='same')
    def composite_noop2():
        noop()

    @composite_solid
    def wrapper():
        composite_noop2()

    @composite_solid
    def top():
        wrapper()
        composite_noop()

    with pytest.raises(DagsterInvalidDefinitionError):

        @pipeline
        def _dupes():
            composite_noop()
            composite_noop2()

    with pytest.raises(DagsterInvalidDefinitionError):
        PipelineDefinition(name='dupes', solid_defs=[top])


def test_two_inputs_with_reversed_input_defs_and_dsl():
    @solid(input_defs=[InputDefinition('num_two'), InputDefinition('num_one')])
    def subtract_ctx(_context, num_one, num_two):
        return num_one - num_two

    @lambda_solid(input_defs=[InputDefinition('num_two'), InputDefinition('num_one')])
    def subtract(num_one, num_two):
        return num_one - num_two

    @pipeline
    def test():
        two = return_two()
        three = return_three()
        subtract(two, three)
        subtract_ctx(two, three)

    assert execute_pipeline(test).result_for_solid('subtract').output_value() == -1
    assert execute_pipeline(test).result_for_solid('subtract_ctx').output_value() == -1


def test_single_non_positional_input_use():
    @lambda_solid(input_defs=[InputDefinition('num')])
    def add_one_kw(**kwargs):
        return kwargs['num'] + 1

    @pipeline
    def test():
        # the decorated solid fn doesn't define args
        # but since there is only one it is unambiguous
        add_one_kw(return_two())

    assert execute_pipeline(test).result_for_solid('add_one_kw').output_value() == 3


def test_single_positional_single_kwarg_input_use():
    @lambda_solid(input_defs=[InputDefinition('num_two'), InputDefinition('num_one')])
    def subtract_kw(num_one, **kwargs):
        return num_one - kwargs['num_two']

    @pipeline
    def test():
        # the decorated solid fn only defines one positional arg
        # and one kwarg so passing two by position is unambiguous
        # since the second argument must be the one kwarg
        subtract_kw(return_two(), return_three())

    assert execute_pipeline(test).result_for_solid('subtract_kw').output_value() == -1


def test_bad_positional_input_use():
    @lambda_solid(
        input_defs=[
            InputDefinition('num_two'),
            InputDefinition('num_one'),
            InputDefinition('num_three'),
        ]
    )
    def add_kw(num_one, **kwargs):
        return num_one + kwargs['num_two'] + kwargs['num_three']

    with pytest.raises(DagsterInvalidDefinitionError, match='Use keyword args instead'):

        @pipeline
        def _fail():
            # the decorated solid fn only defines one positional arg
            # so the two remaining have no positions and this is
            # ambiguous
            add_kw(return_two(), return_two(), return_two())



import pytest

from dagster import DagsterInstance, ModeDefinition, PipelineDefinition, resource, solid
from dagster.core.execution.api import create_execution_plan
from dagster.core.execution.context_creation_pipeline import pipeline_initialization_event_generator
from dagster.core.execution.resources_init import (
    resource_initialization_event_generator,
    resource_initialization_manager,
    single_resource_event_generator,
)
from dagster.core.log_manager import DagsterLogManager
from dagster.core.system_config.objects import EnvironmentConfig
from dagster.core.utils import make_new_run_id


def test_generator_exit():
    def generator():
        try:
            yield 'A'
        finally:
            yield 'EXIT'

    gen = generator()
    next(gen)
    with pytest.raises(RuntimeError, match='generator ignored GeneratorExit'):
        gen.close()


def gen_basic_resource_pipeline(called=None, cleaned=None):
    if not called:
        called = []

    if not cleaned:
        cleaned = []

    @resource
    def resource_a(_):
        try:
            called.append('A')
            yield 'A'
        finally:
            cleaned.append('A')

    @resource
    def resource_b(_):
        try:
            called.append('B')
            yield 'B'
        finally:
            cleaned.append('B')

    @solid(required_resource_keys={'a', 'b'})
    def resource_solid(_):
        pass

    return PipelineDefinition(
        name='basic_resource_pipeline',
        solid_defs=[resource_solid],
        mode_defs=[ModeDefinition(resource_defs={'a': resource_a, 'b': resource_b})],
    )


def test_clean_event_generator_exit():
    ''' Testing for generator cleanup
    (see https://amir.rachum.com/blog/2017/03/03/generator-cleanup/)
    '''
    from dagster.core.execution.context.init import InitResourceContext

    pipeline_def = gen_basic_resource_pipeline()
    instance = DagsterInstance.ephemeral()
    execution_plan = create_execution_plan(pipeline_def)
    pipeline_run = instance.create_run_for_pipeline(
        pipeline_def=pipeline_def, execution_plan=execution_plan
    )
    log_manager = DagsterLogManager(run_id=pipeline_run.run_id, logging_tags={}, loggers=[])
    environment_config = EnvironmentConfig.build(pipeline_def)
    execution_plan = create_execution_plan(pipeline_def)

    resource_name, resource_def = next(iter(pipeline_def.get_default_mode().resource_defs.items()))
    resource_context = InitResourceContext(
        pipeline_def=pipeline_def,
        resource_def=resource_def,
        resource_config=None,
        run_id=make_new_run_id(),
    )
    generator = single_resource_event_generator(resource_context, resource_name, resource_def)
    next(generator)
    generator.close()

    generator = resource_initialization_event_generator(
        execution_plan, environment_config, pipeline_run, log_manager, {'a'}
    )
    next(generator)
    generator.close()

    generator = pipeline_initialization_event_generator(
        execution_plan, {}, pipeline_run, instance, resource_initialization_manager,
    )
    next(generator)
    generator.close()



import sqlalchemy as db

ScheduleStorageSqlMetadata = db.MetaData()

ScheduleTable = db.Table(
    'schedules',
    ScheduleStorageSqlMetadata,
    db.Column('id', db.Integer, primary_key=True, autoincrement=True),
    db.Column('repository_name', db.String(255)),
    db.Column('schedule_name', db.String),
    db.Column('status', db.String(63)),
    db.Column('schedule_body', db.String),
    db.Column('create_timestamp', db.DateTime, server_default=db.text('CURRENT_TIMESTAMP')),
    db.Column('update_timestamp', db.DateTime, server_default=db.text('CURRENT_TIMESTAMP')),
    db.UniqueConstraint('repository_name', 'schedule_name'),
)

ScheduleTickTable = db.Table(
    'schedule_ticks',
    ScheduleStorageSqlMetadata,
    db.Column('id', db.Integer, primary_key=True, autoincrement=True),
    db.Column('repository_name', db.String(255)),
    db.Column('schedule_name', db.String),
    db.Column('status', db.String(63)),
    db.Column('timestamp', db.types.TIMESTAMP),
    db.Column('tick_body', db.String),
    # The create and update timestamps are not used in framework code, are are simply
    # present for debugging purposes.
    db.Column('create_timestamp', db.DateTime, server_default=db.text('CURRENT_TIMESTAMP')),
    db.Column('update_timestamp', db.DateTime, server_default=db.text('CURRENT_TIMESTAMP')),
)



import pytest

from dagster import DagsterInvalidConfigDefinitionError, Noneable, Selector, execute_solid, solid


def test_kitchen_sink():
    @solid(
        config={
            'str_field': str,
            'int_field': int,
            'list_int': [int],
            'list_list_int': [[int]],
            'dict_field': {'a_string': str},
            'list_dict_field': [{'an_int': int}],
            'selector_of_things': Selector(
                {'select_list_dict_field': [{'an_int': int}], 'select_int': int}
            ),
            # this is a good argument to use () instead of [] for type parameterization in
            # the config system
            'optional_list_of_optional_string': Noneable([Noneable(str)]),
        }
    )
    def kitchen_sink(context):
        return context.solid_config

    solid_config_one = {
        'str_field': 'kjf',
        'int_field': 2,
        'list_int': [3],
        'list_list_int': [[1], [2, 3]],
        'dict_field': {'a_string': 'kdjfkd'},
        'list_dict_field': [{'an_int': 2}, {'an_int': 4}],
        'selector_of_things': {'select_int': 3},
        'optional_list_of_optional_string': ['foo', None],
    }

    assert (
        execute_solid(
            kitchen_sink,
            environment_dict={'solids': {'kitchen_sink': {'config': solid_config_one}}},
        ).output_value()
        == solid_config_one
    )

    solid_config_two = {
        'str_field': 'kjf',
        'int_field': 2,
        'list_int': [3],
        'list_list_int': [[1], [2, 3]],
        'dict_field': {'a_string': 'kdjfkd'},
        'list_dict_field': [{'an_int': 2}, {'an_int': 4}],
        'selector_of_things': {'select_list_dict_field': [{'an_int': 5}]},
        'optional_list_of_optional_string': None,
    }

    assert (
        execute_solid(
            kitchen_sink,
            environment_dict={'solids': {'kitchen_sink': {'config': solid_config_two}}},
        ).output_value()
        == solid_config_two
    )


def test_bad_solid_config_argument():
    with pytest.raises(DagsterInvalidConfigDefinitionError) as exc_info:

        @solid(config='dkjfkd')
        def _bad_config(_):
            pass

    assert str(exc_info.value).startswith(
        "Error defining config. Original value passed: 'dkjfkd'. 'dkjfkd' cannot be resolved."
    )


def test_bad_solid_config_argument_nested():
    with pytest.raises(DagsterInvalidConfigDefinitionError) as exc_info:

        @solid(config={'field': 'kdjkfjd'})
        def _bad_config(_):
            pass

    assert str(exc_info.value).startswith(
        "Error defining config. Original value passed: {'field': 'kdjkfjd'}. "
        "Error at stack path :field. 'kdjkfjd' cannot be resolved."
    )


def test_bad_solid_config_argument_list_wrong_length():
    with pytest.raises(DagsterInvalidConfigDefinitionError) as exc_info:

        @solid(config={'bad_list': []})
        def _bad_list_config(_):
            pass

    assert str(exc_info.value).startswith(
        "Error defining config. Original value passed: {'bad_list': []}. "
        "Error at stack path :bad_list. [] cannot be resolved. "
        "Reason: List must be of length 1."
    )


def test_bad_solid_config_argument_list_bad_item():

    with pytest.raises(DagsterInvalidConfigDefinitionError) as exc_info:

        @solid(config={'bad_list': ['kdjfkd']})
        def _bad_list_config(_):
            pass

    assert str(exc_info.value).startswith(
        "Error defining config. Original value passed: {'bad_list': ['kdjfkd']}. "
        "Error at stack path :bad_list. ['kdjfkd'] cannot be resolved. "
        "Reason: List have a single item and contain a valid type i.e. [int]. "
        "Got item 'kdjfkd'."
    )


def test_bad_solid_config_argument_list_bad_nested_item():

    with pytest.raises(DagsterInvalidConfigDefinitionError) as exc_info:

        @solid(config={'bad_nested_list': [{'bad_field': 'kjdkfd'}]})
        def _bad_list_config(_):
            pass

    assert str(exc_info.value).startswith(
        "Error defining config. Original value passed: {'bad_nested_list': "
        "[{'bad_field': 'kjdkfd'}]}. Error at stack path "
        ":bad_nested_list:bad_field. 'kjdkfd' cannot be resolved."
    )



from collections import namedtuple

from dagster import check
from dagster.core.definitions.schedule import ScheduleDefinition, ScheduleExecutionContext
from dagster.core.errors import DagsterInvalidDefinitionError, DagsterInvariantViolationError
from dagster.core.storage.pipeline_run import PipelineRun, PipelineRunStatus, PipelineRunsFilter
from dagster.core.storage.tags import check_tags
from dagster.utils import merge_dicts

from .mode import DEFAULT_MODE_NAME


def by_name(partition):
    return partition.name


class Partition(namedtuple('_Partition', ('value name'))):
    '''
    Partition is the representation of a logical slice across an axis of a pipeline's work

    Args:
        partition (Any): The object for this partition
        name (str): Name for this partition
    '''

    def __new__(cls, value=None, name=None):
        return super(Partition, cls).__new__(
            cls, name=check.opt_str_param(name, 'name', str(value)), value=value
        )


def last_partition(context, partition_set_def):
    check.inst_param(context, 'context', ScheduleExecutionContext)
    partition_set_def = check.inst_param(
        partition_set_def, 'partition_set_def', PartitionSetDefinition
    )

    partitions = partition_set_def.get_partitions()
    if not partitions:
        return None
    return partitions[-1]


def last_empty_partition(context, partition_set_def):
    check.inst_param(context, 'context', ScheduleExecutionContext)
    partition_set_def = check.inst_param(
        partition_set_def, 'partition_set_def', PartitionSetDefinition
    )
    partitions = partition_set_def.get_partitions()
    if not partitions:
        return None
    selected = None
    for partition in reversed(partitions):
        filters = PipelineRunsFilter.for_partition(partition_set_def, partition)
        matching = context.instance.get_runs(filters)
        if not any(run.status == PipelineRunStatus.SUCCESS for run in matching):
            selected = partition
            break
    return selected


def first_partition(context, partition_set_def=None):
    check.inst_param(context, 'context', ScheduleExecutionContext)
    partition_set_def = check.inst_param(
        partition_set_def, 'partition_set_def', PartitionSetDefinition
    )

    partitions = partition_set_def.get_partitions()
    if not partitions:
        return None

    return partitions[0]


class PartitionSetDefinition(
    namedtuple(
        '_PartitionSetDefinition',
        (
            'name pipeline_name partition_fn solid_subset mode user_defined_environment_dict_fn_for_partition user_defined_tags_fn_for_partition'
        ),
    )
):
    '''
    Defines a partition set, representing the set of slices making up an axis of a pipeline

    Args:
        name (str): Name for this partition set
        pipeline_name (str): The name of the pipeline definition
        partition_fn (Callable[void, List[Partition]]): User-provided function to define the set of
            valid partition objects.
        solid_subset (Optional[List[str]]): The list of names of solid invocations (i.e., of
            unaliased solids or of their aliases if aliased) to execute with this partition.
        mode (Optional[str]): The mode to apply when executing this partition. (default: 'default')
        environment_dict_fn_for_partition (Callable[[Partition], [Dict]]): A
            function that takes a Partition and returns the environment
            configuration that parameterizes the execution for this partition, as a dict
        tags_fn_for_partition (Callable[[Partition], Optional[dict[str, str]]]): A function that
            takes a Partition and returns a list of key value pairs that will be
            added to the generated run for this partition.
    '''

    def __new__(
        cls,
        name,
        pipeline_name,
        partition_fn,
        solid_subset=None,
        mode=None,
        environment_dict_fn_for_partition=lambda _partition: {},
        tags_fn_for_partition=lambda _partition: {},
    ):
        def _wrap(x):
            if isinstance(x, Partition):
                return x
            if isinstance(x, str):
                return Partition(x)
            raise DagsterInvalidDefinitionError(
                'Expected <Partition> | <str>, received {type}'.format(type=type(x))
            )

        return super(PartitionSetDefinition, cls).__new__(
            cls,
            name=check.str_param(name, 'name'),
            pipeline_name=check.str_param(pipeline_name, 'pipeline_name'),
            partition_fn=lambda: [
                _wrap(x) for x in check.callable_param(partition_fn, 'partition_fn')()
            ],
            solid_subset=check.opt_nullable_list_param(solid_subset, 'solid_subset', of_type=str),
            mode=check.opt_str_param(mode, 'mode', DEFAULT_MODE_NAME),
            user_defined_environment_dict_fn_for_partition=check.callable_param(
                environment_dict_fn_for_partition, 'environment_dict_fn_for_partition'
            ),
            user_defined_tags_fn_for_partition=check.callable_param(
                tags_fn_for_partition, 'tags_fn_for_partition'
            ),
        )

    def environment_dict_for_partition(self, partition):
        return self.user_defined_environment_dict_fn_for_partition(partition)

    def tags_for_partition(self, partition):
        user_tags = self.user_defined_tags_fn_for_partition(partition)
        check_tags(user_tags, 'user_tags')

        tags = merge_dicts(user_tags, PipelineRun.tags_for_partition_set(self, partition))

        return tags

    def get_partitions(self):
        return self.partition_fn()

    def create_schedule_definition(
        self,
        schedule_name,
        cron_schedule,
        should_execute=None,
        partition_selector=last_partition,
        environment_vars=None,
    ):
        '''Create a ScheduleDefinition from a PartitionSetDefinition.

        Arguments:
            schedule_name (str): The name of the schedule.
            cron_schedule (str): A valid cron string for the schedule
            should_execute (Optional[function]): Function that runs at schedule execution time that
            determines whether a schedule should execute. Defaults to a function that always returns
            ``True``.
            partition_selector (Callable[PartitionSet], Partition): A partition selector for the
                schedule
            environment_vars (Optional[dict]): The environment variables to set for the schedule

        Returns:
            ScheduleDefinition -- The generated ScheduleDefinition for the partition selector
        '''

        check.str_param(schedule_name, 'schedule_name')
        check.str_param(cron_schedule, 'cron_schedule')
        check.opt_callable_param(should_execute, 'should_execute')
        check.opt_dict_param(environment_vars, 'environment_vars', key_type=str, value_type=str)
        check.callable_param(partition_selector, 'partition_selector')

        def _should_execute_wrapper(context):
            check.inst_param(context, 'context', ScheduleExecutionContext)
            selected_partition = partition_selector(context, self)
            if not selected_partition:
                return False
            elif not should_execute:
                return True
            else:
                return should_execute(context)

        def _environment_dict_fn_wrapper(context):
            check.inst_param(context, 'context', ScheduleExecutionContext)
            selected_partition = partition_selector(context, self)
            if not selected_partition:
                raise DagsterInvariantViolationError(
                    "The partition selection function `{selector}` did not return "
                    "a partition from PartitionSet {partition_set}".format(
                        selector=getattr(partition_selector, '__name__', repr(partition_selector)),
                        partition_set=self.name,
                    )
                )

            return self.environment_dict_for_partition(selected_partition)

        def _tags_fn_wrapper(context):
            check.inst_param(context, 'context', ScheduleExecutionContext)
            selected_partition = partition_selector(context, self)
            if not selected_partition:
                raise DagsterInvariantViolationError(
                    "The partition selection function `{selector}` did not return "
                    "a partition from PartitionSet {partition_set}".format(
                        selector=getattr(partition_selector, '__name__', repr(partition_selector)),
                        partition_set=self.name,
                    )
                )

            return self.tags_for_partition(selected_partition)

        return PartitionScheduleDefinition(
            name=schedule_name,
            cron_schedule=cron_schedule,
            pipeline_name=self.pipeline_name,
            environment_dict_fn=_environment_dict_fn_wrapper,
            tags_fn=_tags_fn_wrapper,
            solid_subset=self.solid_subset,
            mode=self.mode,
            should_execute=_should_execute_wrapper,
            environment_vars=environment_vars,
            partition_set=self,
        )


class PartitionScheduleDefinition(ScheduleDefinition):
    __slots__ = ['_partition_set']

    def __init__(
        self,
        name,
        cron_schedule,
        pipeline_name,
        environment_dict_fn,
        tags_fn,
        solid_subset,
        mode,
        should_execute,
        environment_vars,
        partition_set,
    ):
        super(PartitionScheduleDefinition, self).__init__(
            name=name,
            cron_schedule=cron_schedule,
            pipeline_name=pipeline_name,
            environment_dict_fn=environment_dict_fn,
            tags_fn=tags_fn,
            solid_subset=solid_subset,
            mode=mode,
            should_execute=should_execute,
            environment_vars=environment_vars,
        )
        self._partition_set = check.inst_param(
            partition_set, 'partition_set', PartitionSetDefinition
        )

    def get_partition_set(self):
        return self._partition_set


class RepositoryPartitionsHandle(namedtuple('_RepositoryPartitionsHandle', ('partition_set_defs'))):
    def __new__(cls, partition_set_defs):
        return super(RepositoryPartitionsHandle, cls).__new__(
            cls,
            partition_set_defs=check.list_param(
                partition_set_defs, 'partition_set_defs', PartitionSetDefinition
            ),
        )

    def get_partition_sets(self):
        return self.partition_set_defs

    def get_partition_set(self, name):
        return next(
            iter(
                _partition_set
                for _partition_set in self.partition_set_defs
                if _partition_set.name == name
            ),
            None,
        )


def repository_partitions(arg=None):
    ''' Decorator to define the set of partition sets for a repository '''
    if callable(arg):
        return RepositoryPartitionsHandle(partition_set_defs=arg())

    def _wrap(fn):
        return RepositoryPartitionsHandle(partition_set_defs=fn())

    return _wrap



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots['test_create_noop_execution_plan 1'] = '''{
  "__class__": "ExecutionPlanSnapshot",
  "artifacts_persisted": false,
  "pipeline_snapshot_id": "e95e6f29f25ea236ce191a1e8b49ae8601a0afef",
  "steps": [
    {
      "__class__": "ExecutionStepSnap",
      "inputs": [],
      "key": "noop_solid.compute",
      "kind": {
        "__enum__": "StepKind.COMPUTE"
      },
      "metadata_items": [],
      "outputs": [
        {
          "__class__": "ExecutionStepOutputSnap",
          "dagster_type_key": "Any",
          "name": "result"
        }
      ],
      "solid_handle_id": "noop_solid"
    }
  ]
}'''

snapshots['test_create_execution_plan_with_dep 1'] = '''{
  "__class__": "ExecutionPlanSnapshot",
  "artifacts_persisted": false,
  "pipeline_snapshot_id": "1f30f3ee72df05cdda75a455f47f525dc9f9b7d7",
  "steps": [
    {
      "__class__": "ExecutionStepSnap",
      "inputs": [],
      "key": "solid_one.compute",
      "kind": {
        "__enum__": "StepKind.COMPUTE"
      },
      "metadata_items": [],
      "outputs": [
        {
          "__class__": "ExecutionStepOutputSnap",
          "dagster_type_key": "Any",
          "name": "result"
        }
      ],
      "solid_handle_id": "solid_one"
    },
    {
      "__class__": "ExecutionStepSnap",
      "inputs": [
        {
          "__class__": "ExecutionStepInputSnap",
          "dagster_type_key": "Any",
          "name": "num",
          "upstream_output_handles": [
            {
              "__class__": "StepOutputHandle",
              "output_name": "result",
              "step_key": "solid_one.compute"
            }
          ]
        }
      ],
      "key": "solid_two.compute",
      "kind": {
        "__enum__": "StepKind.COMPUTE"
      },
      "metadata_items": [],
      "outputs": [
        {
          "__class__": "ExecutionStepOutputSnap",
          "dagster_type_key": "Any",
          "name": "result"
        }
      ],
      "solid_handle_id": "solid_two"
    }
  ]
}'''

snapshots['test_create_with_composite 1'] = '''{
  "__class__": "ExecutionPlanSnapshot",
  "artifacts_persisted": false,
  "pipeline_snapshot_id": "c702d45a160714c38fcbde6427dabfc38aca1ab3",
  "steps": [
    {
      "__class__": "ExecutionStepSnap",
      "inputs": [
        {
          "__class__": "ExecutionStepInputSnap",
          "dagster_type_key": "Any",
          "name": "num_one",
          "upstream_output_handles": [
            {
              "__class__": "StepOutputHandle",
              "output_name": "result",
              "step_key": "comp_1.add_one.compute"
            }
          ]
        },
        {
          "__class__": "ExecutionStepInputSnap",
          "dagster_type_key": "Any",
          "name": "num_two",
          "upstream_output_handles": [
            {
              "__class__": "StepOutputHandle",
              "output_name": "result",
              "step_key": "comp_2.add_one.compute"
            }
          ]
        }
      ],
      "key": "add.compute",
      "kind": {
        "__enum__": "StepKind.COMPUTE"
      },
      "metadata_items": [],
      "outputs": [
        {
          "__class__": "ExecutionStepOutputSnap",
          "dagster_type_key": "Any",
          "name": "result"
        }
      ],
      "solid_handle_id": "add"
    },
    {
      "__class__": "ExecutionStepSnap",
      "inputs": [
        {
          "__class__": "ExecutionStepInputSnap",
          "dagster_type_key": "Int",
          "name": "num",
          "upstream_output_handles": [
            {
              "__class__": "StepOutputHandle",
              "output_name": "out_num",
              "step_key": "comp_1.return_one.compute"
            }
          ]
        }
      ],
      "key": "comp_1.add_one.compute",
      "kind": {
        "__enum__": "StepKind.COMPUTE"
      },
      "metadata_items": [],
      "outputs": [
        {
          "__class__": "ExecutionStepOutputSnap",
          "dagster_type_key": "Int",
          "name": "result"
        }
      ],
      "solid_handle_id": "comp_1.add_one"
    },
    {
      "__class__": "ExecutionStepSnap",
      "inputs": [],
      "key": "comp_1.return_one.compute",
      "kind": {
        "__enum__": "StepKind.COMPUTE"
      },
      "metadata_items": [],
      "outputs": [
        {
          "__class__": "ExecutionStepOutputSnap",
          "dagster_type_key": "Int",
          "name": "out_num"
        }
      ],
      "solid_handle_id": "comp_1.return_one"
    },
    {
      "__class__": "ExecutionStepSnap",
      "inputs": [
        {
          "__class__": "ExecutionStepInputSnap",
          "dagster_type_key": "Int",
          "name": "num",
          "upstream_output_handles": [
            {
              "__class__": "StepOutputHandle",
              "output_name": "out_num",
              "step_key": "comp_2.return_one.compute"
            }
          ]
        }
      ],
      "key": "comp_2.add_one.compute",
      "kind": {
        "__enum__": "StepKind.COMPUTE"
      },
      "metadata_items": [],
      "outputs": [
        {
          "__class__": "ExecutionStepOutputSnap",
          "dagster_type_key": "Int",
          "name": "result"
        }
      ],
      "solid_handle_id": "comp_2.add_one"
    },
    {
      "__class__": "ExecutionStepSnap",
      "inputs": [],
      "key": "comp_2.return_one.compute",
      "kind": {
        "__enum__": "StepKind.COMPUTE"
      },
      "metadata_items": [],
      "outputs": [
        {
          "__class__": "ExecutionStepOutputSnap",
          "dagster_type_key": "Int",
          "name": "out_num"
        }
      ],
      "solid_handle_id": "comp_2.return_one"
    }
  ]
}'''

snapshots['test_create_noop_execution_plan_with_tags 1'] = '''{
  "__class__": "ExecutionPlanSnapshot",
  "artifacts_persisted": false,
  "pipeline_snapshot_id": "be0ce8dd36ee2102a34795cc907ec48d10951f02",
  "steps": [
    {
      "__class__": "ExecutionStepSnap",
      "inputs": [],
      "key": "noop_solid.compute",
      "kind": {
        "__enum__": "StepKind.COMPUTE"
      },
      "metadata_items": [
        {
          "__class__": "ExecutionPlanMetadataItemSnap",
          "key": "bar",
          "value": "baaz"
        },
        {
          "__class__": "ExecutionPlanMetadataItemSnap",
          "key": "foo",
          "value": "bar"
        }
      ],
      "outputs": [
        {
          "__class__": "ExecutionStepOutputSnap",
          "dagster_type_key": "Any",
          "name": "result"
        }
      ],
      "solid_handle_id": "noop_solid"
    }
  ]
}'''



import time

import pytest

from dagster import DagsterInvariantViolationError, RepositoryDefinition
from dagster.core.scheduler import Schedule, ScheduleDefinitionData, ScheduleStatus
from dagster.core.scheduler.scheduler import ScheduleTickData, ScheduleTickStatus
from dagster.utils.error import SerializableErrorInfo


class TestScheduleStorage:
    '''
    You can extend this class to easily run these set of tests on any schedule storage. When extending,
    you simply need to override the `schedule_storage` fixture and return your implementation of
    `ScheduleStorage`.

    For example:

    ```
    TestScheduleStorage.__test__ = False

    class TestMyStorageImplementation(TestScheduleStorage):
        __test__ = True

        @pytest.fixture(scope='function', name='storage')
        def schedule_storage(self):  # pylint: disable=arguments-differ
            return MyStorageImplementation()
    ```
    '''

    @pytest.fixture(name='storage', params=[])
    def schedule_storage(self, request):
        with request.param() as s:
            yield s

    @staticmethod
    def build_schedule(
        schedule_name, cron_schedule, status=ScheduleStatus.STOPPED,
    ):
        return Schedule(ScheduleDefinitionData(schedule_name, cron_schedule), status, "", "",)

    def test_basic_schedule_storage(self, storage):
        assert storage

        repository = RepositoryDefinition("repository_name")
        schedule = self.build_schedule("my_schedule", "* * * * *")
        storage.add_schedule(repository, schedule)
        schedules = storage.all_schedules(repository)
        assert len(schedules) == 1

        schedule = schedules[0]
        assert schedule.name == "my_schedule"
        assert schedule.cron_schedule == "* * * * *"

    def test_add_multiple_schedules(self, storage):
        assert storage

        repository = RepositoryDefinition("repository_name")
        schedule = self.build_schedule("my_schedule", "* * * * *")
        schedule_2 = self.build_schedule("my_schedule_2", "* * * * *")
        schedule_3 = self.build_schedule("my_schedule_3", "* * * * *")

        storage.add_schedule(repository, schedule)
        storage.add_schedule(repository, schedule_2)
        storage.add_schedule(repository, schedule_3)

        schedules = storage.all_schedules(repository)
        assert len(schedules) == 3

        assert any(s.name == "my_schedule" for s in schedules)
        assert any(s.name == "my_schedule_2" for s in schedules)
        assert any(s.name == "my_schedule_3" for s in schedules)

    def test_get_schedule_by_name(self, storage):
        assert storage

        repository = RepositoryDefinition("repository_name")
        storage.add_schedule(repository, self.build_schedule("my_schedule", "* * * * *"))
        schedule = storage.get_schedule_by_name(repository, "my_schedule")

        assert schedule.name == "my_schedule"

    def test_get_schedule_by_name_not_found(self, storage):
        assert storage

        repository = RepositoryDefinition("repository_name")
        storage.add_schedule(repository, self.build_schedule("my_schedule", "* * * * *"))
        schedule = storage.get_schedule_by_name(repository, "fake_schedule")

        assert schedule is None

    def test_update_schedule(self, storage):
        assert storage

        repository = RepositoryDefinition("repository_name")
        schedule = self.build_schedule("my_schedule", "* * * * *")
        storage.add_schedule(repository, schedule)

        new_schedule = schedule.with_status(ScheduleStatus.RUNNING)
        storage.update_schedule(repository, new_schedule)

        schedules = storage.all_schedules(repository)
        assert len(schedules) == 1

        schedule = schedules[0]
        assert schedule.name == "my_schedule"
        assert schedule.status == ScheduleStatus.RUNNING

    def test_update_schedule_not_found(self, storage):
        assert storage

        repository = RepositoryDefinition("repository_name")
        schedule = self.build_schedule("my_schedule", "* * * * *")

        with pytest.raises(DagsterInvariantViolationError):
            storage.update_schedule(repository, schedule)

    def test_delete_schedule(self, storage):
        assert storage

        repository = RepositoryDefinition("repository_name")
        schedule = self.build_schedule("my_schedule", "* * * * *")
        storage.add_schedule(repository, schedule)
        storage.delete_schedule(repository, schedule)

        schedules = storage.all_schedules(repository)
        assert len(schedules) == 0

    def test_delete_schedule_not_found(self, storage):
        assert storage

        repository = RepositoryDefinition("repository_name")
        schedule = self.build_schedule("my_schedule", "* * * * *")

        with pytest.raises(DagsterInvariantViolationError):
            storage.delete_schedule(repository, schedule)

    def test_add_schedule_with_same_name(self, storage):
        assert storage

        repository = RepositoryDefinition("repository_name")
        schedule = self.build_schedule("my_schedule", "* * * * *")
        storage.add_schedule(repository, schedule)

        with pytest.raises(DagsterInvariantViolationError):
            storage.add_schedule(repository, schedule)

    def build_tick(self, current_time, status=ScheduleTickStatus.STARTED, run_id=None, error=None):
        return ScheduleTickData("my_schedule", "* * * * *", current_time, status, run_id, error)

    def test_create_tick(self, storage):
        assert storage

        repository = RepositoryDefinition("repository_name")
        current_time = time.time()
        tick = storage.create_schedule_tick(repository, self.build_tick(current_time))
        assert tick.tick_id == 1

        ticks = storage.get_schedule_ticks_by_schedule(repository, "my_schedule")
        assert len(ticks) == 1
        tick = ticks[0]
        assert tick.tick_id == 1
        assert tick.schedule_name == "my_schedule"
        assert tick.cron_schedule == "* * * * *"
        assert tick.timestamp == current_time
        assert tick.status == ScheduleTickStatus.STARTED
        assert tick.run_id == None
        assert tick.error == None

    def test_update_tick_to_success(self, storage):
        assert storage

        repository = RepositoryDefinition("repository_name")
        current_time = time.time()
        tick = storage.create_schedule_tick(repository, self.build_tick(current_time))

        updated_tick = tick.with_status(ScheduleTickStatus.SUCCESS, run_id="1234")
        assert updated_tick.status == ScheduleTickStatus.SUCCESS

        storage.update_schedule_tick(repository, updated_tick)

        ticks = storage.get_schedule_ticks_by_schedule(repository, "my_schedule")
        assert len(ticks) == 1
        tick = ticks[0]
        assert tick.tick_id == 1
        assert tick.schedule_name == "my_schedule"
        assert tick.cron_schedule == "* * * * *"
        assert tick.timestamp == current_time
        assert tick.status == ScheduleTickStatus.SUCCESS
        assert tick.run_id == "1234"
        assert tick.error == None

    def test_update_tick_to_skip(self, storage):
        assert storage

        repository = RepositoryDefinition("repository_name")
        current_time = time.time()
        tick = storage.create_schedule_tick(repository, self.build_tick(current_time))

        updated_tick = tick.with_status(ScheduleTickStatus.SKIPPED)
        assert updated_tick.status == ScheduleTickStatus.SKIPPED

        storage.update_schedule_tick(repository, updated_tick)

        ticks = storage.get_schedule_ticks_by_schedule(repository, "my_schedule")
        assert len(ticks) == 1
        tick = ticks[0]
        assert tick.tick_id == 1
        assert tick.schedule_name == "my_schedule"
        assert tick.cron_schedule == "* * * * *"
        assert tick.timestamp == current_time
        assert tick.status == ScheduleTickStatus.SKIPPED
        assert tick.run_id == None
        assert tick.error == None

    def test_update_tick_to_failure(self, storage):
        assert storage

        repository = RepositoryDefinition("repository_name")
        current_time = time.time()
        tick = storage.create_schedule_tick(repository, self.build_tick(current_time))

        updated_tick = tick.with_status(
            ScheduleTickStatus.FAILURE,
            error=SerializableErrorInfo(message="Error", stack=[], cls_name="TestError"),
        )
        assert updated_tick.status == ScheduleTickStatus.FAILURE

        storage.update_schedule_tick(repository, updated_tick)

        ticks = storage.get_schedule_ticks_by_schedule(repository, "my_schedule")
        assert len(ticks) == 1
        tick = ticks[0]
        assert tick.tick_id == 1
        assert tick.schedule_name == "my_schedule"
        assert tick.cron_schedule == "* * * * *"
        assert tick.timestamp == current_time
        assert tick.status == ScheduleTickStatus.FAILURE
        assert tick.run_id == None
        assert tick.error == SerializableErrorInfo(message="Error", stack=[], cls_name="TestError")

    def test_get_schedule_stats(self, storage):
        assert storage

        repository = RepositoryDefinition("repository_name")
        current_time = time.time()

        error = SerializableErrorInfo(message="Error", stack=[], cls_name="TestError")

        # Create ticks
        for x in range(2):
            storage.create_schedule_tick(repository, self.build_tick(current_time))

        for x in range(3):
            storage.create_schedule_tick(
                repository, self.build_tick(current_time, ScheduleTickStatus.SUCCESS, run_id=str(x))
            )

        for x in range(4):
            storage.create_schedule_tick(
                repository, self.build_tick(current_time, ScheduleTickStatus.SKIPPED),
            )

        for x in range(5):
            storage.create_schedule_tick(
                repository, self.build_tick(current_time, ScheduleTickStatus.FAILURE, error=error),
            )

        stats = storage.get_schedule_tick_stats_by_schedule(repository, "my_schedule")
        assert stats.ticks_started == 2
        assert stats.ticks_succeeded == 3
        assert stats.ticks_skipped == 4
        assert stats.ticks_failed == 5



from dagster import InputDefinition, List, Optional, execute_pipeline, lambda_solid, pipeline


def test_from_intermediates_from_multiple_outputs():
    @lambda_solid
    def x():
        return 'x'

    @lambda_solid
    def y():
        return 'y'

    @lambda_solid(input_defs=[InputDefinition('stuff', Optional[List[str]])])
    def gather(stuff):
        return "{} and {}".format(*stuff)

    @pipeline
    def pipe():
        gather([x(), y()])

    result = execute_pipeline(pipe)

    assert result
    assert result.success
    assert (
        result.result_for_solid('gather')
        .compute_input_event_dict['stuff']
        .event_specific_data[1]
        .label
        == 'stuff'
    )
    assert result.result_for_solid('gather').output_value() == 'x and y'


def test_from_intermediates_from_config():
    environment_dict = {
        'solids': {'x': {'inputs': {'string_input': {'value': 'Dagster is great!'}}}}
    }

    @lambda_solid
    def x(string_input):
        return string_input

    @pipeline
    def pipe():
        x()

    result = execute_pipeline(pipe, environment_dict=environment_dict)

    assert result
    assert result.success
    assert (
        result.result_for_solid('x')
        .compute_input_event_dict['string_input']
        .event_specific_data[1]
        .label
        == 'string_input'
    )
    assert result.result_for_solid('x').output_value() == 'Dagster is great!'



import re

import pytest

from dagster import (
    Bool,
    DagsterInvalidDefinitionError,
    DagsterInvariantViolationError,
    ModeDefinition,
    PipelineDefinition,
    PresetDefinition,
    check,
    execute_pipeline,
    lambda_solid,
    solid,
)
from dagster.utils import file_relative_path


def test_presets():
    @solid(config={'error': Bool})
    def can_fail(context):
        if context.solid_config['error']:
            raise Exception('I did an error')
        return 'cool'

    @lambda_solid
    def always_fail():
        raise Exception('I always do this')

    pipeline = PipelineDefinition(
        name='simple',
        solid_defs=[can_fail, always_fail],
        preset_defs=[
            PresetDefinition.from_files(
                'passing',
                environment_files=[file_relative_path(__file__, 'pass_env.yaml')],
                solid_subset=['can_fail'],
            ),
            PresetDefinition.from_files(
                'passing_overide_to_fail',
                environment_files=[file_relative_path(__file__, 'pass_env.yaml')],
                solid_subset=['can_fail'],
            ).with_additional_config({'solids': {'can_fail': {'config': {'error': True}}}}),
            PresetDefinition(
                'passing_direct_dict',
                environment_dict={'solids': {'can_fail': {'config': {'error': False}}}},
                solid_subset=['can_fail'],
            ),
            PresetDefinition.from_files(
                'failing_1',
                environment_files=[file_relative_path(__file__, 'fail_env.yaml')],
                solid_subset=['can_fail'],
            ),
            PresetDefinition.from_files(
                'failing_2', environment_files=[file_relative_path(__file__, 'pass_env.yaml')]
            ),
            PresetDefinition('subset', solid_subset=['can_fail'],),
        ],
    )

    with pytest.raises(DagsterInvalidDefinitionError):
        PresetDefinition.from_files(
            'invalid_1', environment_files=[file_relative_path(__file__, 'not_a_file.yaml')]
        )

    with pytest.raises(DagsterInvariantViolationError):
        PresetDefinition.from_files(
            'invalid_2',
            environment_files=[file_relative_path(__file__, 'test_repository_definition.py')],
        )

    assert execute_pipeline(pipeline, preset='passing').success

    assert execute_pipeline(pipeline, preset='passing_direct_dict').success
    assert execute_pipeline(pipeline, preset='failing_1', raise_on_error=False).success == False

    assert execute_pipeline(pipeline, preset='failing_2', raise_on_error=False).success == False

    with pytest.raises(DagsterInvariantViolationError, match='Could not find preset'):
        execute_pipeline(pipeline, preset='not_failing', raise_on_error=False)

    assert (
        execute_pipeline(pipeline, preset='passing_overide_to_fail', raise_on_error=False).success
        == False
    )

    assert execute_pipeline(
        pipeline,
        preset='passing',
        environment_dict={'solids': {'can_fail': {'config': {'error': False}}}},
    ).success

    with pytest.raises(
        check.CheckError,
        match=re.escape(
            'The environment set in preset \'passing\' does not agree with the environment passed '
            'in the `environment_dict` argument.'
        ),
    ):
        execute_pipeline(
            pipeline,
            preset='passing',
            environment_dict={'solids': {'can_fail': {'config': {'error': True}}}},
        )

    assert execute_pipeline(
        pipeline,
        preset='subset',
        environment_dict={'solids': {'can_fail': {'config': {'error': False}}}},
    ).success


def test_invalid_preset():
    @lambda_solid
    def lil_solid():
        return ';)'

    with pytest.raises(DagsterInvalidDefinitionError, match='mode "mode_b" which is not defined'):
        PipelineDefinition(
            name="preset_modes",
            solid_defs=[lil_solid],
            mode_defs=[ModeDefinition(name="mode_a")],
            preset_defs=[PresetDefinition(name="preset_b", mode="mode_b")],
        )


def test_from_yaml_strings():
    a = '''
foo:
  bar: 1
baz: 3
'''
    b = '''
foo:
  one: "one"
other: 4
'''
    c = '''
final: "result"
'''
    preset = PresetDefinition.from_yaml_strings('passing', [a, b, c])
    assert preset.environment_dict == {
        'foo': {'bar': 1, 'one': 'one'},
        'baz': 3,
        'other': 4,
        'final': 'result',
    }
    with pytest.raises(
        DagsterInvariantViolationError, match='Encountered error attempting to parse yaml'
    ):
        PresetDefinition.from_yaml_strings('failing', ['--- `'])

    res = PresetDefinition.from_yaml_strings('empty')
    assert res == PresetDefinition(
        name='empty', environment_dict={}, solid_subset=None, mode='default'
    )


def test_from_pkg_resources():
    good = ('dagster_tests.core_tests.definitions_tests', 'pass_env.yaml')
    res = PresetDefinition.from_pkg_resources('pass', [good])
    assert res.environment_dict == {'solids': {'can_fail': {'config': {'error': False}}}}

    bad_defs = [
        ('dagster_tests.core_tests.definitions_tests', 'does_not_exist.yaml'),
        ('dagster_tests.core_tests.definitions_tests', 'bad_file_binary.yaml'),
        ('dagster_tests.core_tests.does_not_exist', 'some_file.yaml'),
    ]

    for bad_def in bad_defs:
        with pytest.raises(
            DagsterInvariantViolationError, match='Encountered error attempting to parse yaml',
        ):
            PresetDefinition.from_pkg_resources('bad_def', [bad_def])



import re

import mock
import pytest

from dagster import OutputDefinition
from dagster.check import CheckError


def test_output_definition_is_optional_field():
    with mock.patch('warnings.warn'):
        output_definition_onlyopt = OutputDefinition(
            dagster_type=int, name='result', is_optional=True
        )
        assert output_definition_onlyopt.optional is True

    with pytest.raises(
        CheckError,
        match=re.escape('Do not use deprecated "is_optional" now that you are using "is_required"'),
    ):
        OutputDefinition(dagster_type=int, name='result', is_optional=True, is_required=True)


def test_output_definition():
    output_definiton_onlyreq = OutputDefinition(dagster_type=int, name='result', is_required=True)
    assert output_definiton_onlyreq.optional is False

    output_definiton_none = OutputDefinition(dagster_type=int, name='result')
    assert output_definiton_none.optional is False



from __future__ import print_function

import os
import string
from contextlib import contextmanager

import mock
import pytest
from click import UsageError
from click.testing import CliRunner

from dagster import (
    DagsterInvariantViolationError,
    PartitionSetDefinition,
    RepositoryDefinition,
    ScheduleDefinition,
    lambda_solid,
    pipeline,
    seven,
    solid,
)
from dagster.check import CheckError
from dagster.cli.pipeline import (
    execute_backfill_command,
    execute_execute_command,
    execute_list_command,
    execute_print_command,
    execute_scaffold_command,
    pipeline_backfill_command,
    pipeline_execute_command,
    pipeline_launch_command,
    pipeline_list_command,
    pipeline_print_command,
    pipeline_scaffold_command,
)
from dagster.cli.run import run_list_command, run_wipe_command
from dagster.cli.schedule import (
    schedule_list_command,
    schedule_restart_command,
    schedule_start_command,
    schedule_stop_command,
    schedule_up_command,
    schedule_wipe_command,
)
from dagster.config.field_utils import Shape
from dagster.core.instance import DagsterInstance, InstanceType
from dagster.core.launcher import RunLauncher
from dagster.core.storage.event_log import InMemoryEventLogStorage
from dagster.core.storage.local_compute_log_manager import NoOpComputeLogManager
from dagster.core.storage.root import LocalArtifactStorage
from dagster.core.storage.runs import InMemoryRunStorage
from dagster.core.storage.schedules import SqliteScheduleStorage
from dagster.serdes import ConfigurableClass
from dagster.utils import file_relative_path
from dagster.utils.test import FilesytemTestScheduler


def no_print(_):
    return None


@lambda_solid
def do_something():
    return 1


@lambda_solid
def do_input(x):
    return x


@pipeline(name='foo')
def foo_pipeline():
    do_input(do_something())


def define_foo_pipeline():
    return foo_pipeline


@pipeline(name='baz', description='Not much tbh')
def baz_pipeline():
    do_input()


def define_bar_repo():
    return RepositoryDefinition(
        'bar',
        {'foo': define_foo_pipeline, 'baz': lambda: baz_pipeline},
        schedule_defs=define_bar_schedules(),
        partition_set_defs=define_baz_partitions(),
    )


@solid
def spew(context):
    context.log.info('HELLO WORLD')


@solid
def fail(context):
    raise Exception('I AM SUPPOSED TO FAIL')


@pipeline
def stdout_pipeline():
    spew()


@pipeline
def stderr_pipeline():
    fail()


def test_list_command():
    runner = CliRunner()

    execute_list_command(
        {
            'repository_yaml': None,
            'python_file': file_relative_path(__file__, 'test_cli_commands.py'),
            'module_name': None,
            'fn_name': 'define_bar_repo',
        },
        no_print,
    )

    result = runner.invoke(
        pipeline_list_command,
        ['-f', file_relative_path(__file__, 'test_cli_commands.py'), '-n', 'define_bar_repo'],
    )

    assert result.exit_code == 0
    assert result.output == (
        'Repository bar\n'
        '**************\n'
        'Pipeline: baz\n'
        'Description:\n'
        'Not much tbh\n'
        'Solids: (Execution Order)\n'
        '    do_input\n'
        '*************\n'
        'Pipeline: foo\n'
        'Solids: (Execution Order)\n'
        '    do_something\n'
        '    do_input\n'
    )

    execute_list_command(
        {
            'repository_yaml': None,
            'python_file': None,
            'module_name': 'dagster_examples.intro_tutorial.repos',
            'fn_name': 'define_repo',
        },
        no_print,
    )

    result = runner.invoke(
        pipeline_list_command, ['-m', 'dagster_examples.intro_tutorial.repos', '-n', 'define_repo']
    )
    assert result.exit_code == 0
    assert result.output == (
        'Repository hello_cereal_repository\n'
        '**********************************\n'
        'Pipeline: complex_pipeline\n'
        'Solids: (Execution Order)\n'
        '    load_cereals\n'
        '    sort_by_calories\n'
        '    sort_by_protein\n'
        '    display_results\n'
        '*******************************\n'
        'Pipeline: hello_cereal_pipeline\n'
        'Solids: (Execution Order)\n'
        '    hello_cereal\n'
    )

    execute_list_command(
        {
            'repository_yaml': file_relative_path(__file__, 'repository_module.yaml'),
            'python_file': None,
            'module_name': None,
            'fn_name': None,
        },
        no_print,
    )

    result = runner.invoke(
        pipeline_list_command, ['-y', file_relative_path(__file__, 'repository_module.yaml')]
    )
    assert result.exit_code == 0
    assert result.output == (
        'Repository hello_cereal_repository\n'
        '**********************************\n'
        'Pipeline: complex_pipeline\n'
        'Solids: (Execution Order)\n'
        '    load_cereals\n'
        '    sort_by_calories\n'
        '    sort_by_protein\n'
        '    display_results\n'
        '*******************************\n'
        'Pipeline: hello_cereal_pipeline\n'
        'Solids: (Execution Order)\n'
        '    hello_cereal\n'
    )

    with pytest.raises(UsageError):
        execute_list_command(
            {
                'repository_yaml': None,
                'python_file': 'foo.py',
                'module_name': 'dagster_examples.intro_tutorial.repos',
                'fn_name': 'define_repo',
            },
            no_print,
        )

    result = runner.invoke(
        pipeline_list_command,
        ['-f', 'foo.py', '-m', 'dagster_examples.intro_tutorial.repos', '-n', 'define_repo'],
    )
    assert result.exit_code == 2

    with pytest.raises(UsageError):
        execute_list_command(
            {
                'repository_yaml': None,
                'python_file': None,
                'module_name': 'dagster_examples.intro_tutorial.repos',
                'fn_name': None,
            },
            no_print,
        )

    result = runner.invoke(pipeline_list_command, ['-m', 'dagster_examples.intro_tutorial.repos'])
    assert result.exit_code == 2

    with pytest.raises(UsageError):
        execute_list_command(
            {
                'repository_yaml': None,
                'python_file': file_relative_path(__file__, 'test_cli_commands.py'),
                'module_name': None,
                'fn_name': None,
            },
            no_print,
        )

    result = runner.invoke(
        pipeline_list_command, ['-f', file_relative_path(__file__, 'test_cli_commands.py')]
    )
    assert result.exit_code == 2


def valid_execute_args():
    return [
        {
            'repository_yaml': file_relative_path(__file__, 'repository_file.yaml'),
            'pipeline_name': ('foo',),
            'python_file': None,
            'module_name': None,
            'fn_name': None,
        },
        {
            'repository_yaml': file_relative_path(__file__, 'repository_module.yaml'),
            'pipeline_name': ('hello_cereal_pipeline',),
            'python_file': None,
            'module_name': None,
            'fn_name': None,
        },
        {
            'repository_yaml': None,
            'pipeline_name': ('foo',),
            'python_file': file_relative_path(__file__, 'test_cli_commands.py'),
            'module_name': None,
            'fn_name': 'define_bar_repo',
        },
        {
            'repository_yaml': None,
            'pipeline_name': ('hello_cereal_pipeline',),
            'python_file': None,
            'module_name': 'dagster_examples.intro_tutorial.repos',
            'fn_name': 'define_repo',
        },
        {
            'repository_yaml': None,
            'pipeline_name': (),
            'python_file': None,
            'module_name': 'dagster_examples.intro_tutorial.repos',
            'fn_name': 'hello_cereal_pipeline',
        },
        {
            'repository_yaml': None,
            'pipeline_name': (),
            'python_file': file_relative_path(__file__, 'test_cli_commands.py'),
            'module_name': None,
            'fn_name': 'define_foo_pipeline',
        },
        {
            'repository_yaml': None,
            'pipeline_name': (),
            'python_file': file_relative_path(__file__, 'test_cli_commands.py'),
            'module_name': None,
            'fn_name': 'foo_pipeline',
        },
    ]


def valid_cli_args():
    return [
        ['-y', file_relative_path(__file__, 'repository_file.yaml'), 'foo'],
        ['-y', file_relative_path(__file__, 'repository_module.yaml'), 'hello_cereal_pipeline'],
        [
            '-f',
            file_relative_path(__file__, 'test_cli_commands.py'),
            '-n',
            'define_bar_repo',
            'foo',
        ],
        [
            '-m',
            'dagster_examples.intro_tutorial.repos',
            '-n',
            'define_repo',
            'hello_cereal_pipeline',
        ],
        ['-m', 'dagster_examples.intro_tutorial.repos', '-n', 'hello_cereal_pipeline'],
        ['-f', file_relative_path(__file__, 'test_cli_commands.py'), '-n', 'define_foo_pipeline'],
    ]


def test_print_command():
    for cli_args in valid_execute_args():
        execute_print_command(verbose=True, cli_args=cli_args, print_fn=no_print)

    for cli_args in valid_execute_args():
        execute_print_command(verbose=False, cli_args=cli_args, print_fn=no_print)

    runner = CliRunner()

    for cli_args in valid_cli_args():
        result = runner.invoke(pipeline_print_command, cli_args)
        assert result.exit_code == 0

        result = runner.invoke(pipeline_print_command, ['--verbose'] + cli_args)
        assert result.exit_code == 0

    res = runner.invoke(
        pipeline_print_command,
        [
            '--verbose',
            '-f',
            file_relative_path(__file__, 'test_cli_commands.py'),
            '-n',
            'define_bar_repo',
            'baz',
        ],
    )
    assert res.exit_code == 0


def test_execute_mode_command():
    runner = CliRunner()

    add_result = runner_pipeline_execute(
        runner,
        [
            '-y',
            file_relative_path(__file__, '../repository.yaml'),
            '--env',
            file_relative_path(__file__, '../environments/multi_mode_with_resources/add_mode.yaml'),
            '-d',
            'add_mode',
            'multi_mode_with_resources',  # pipeline name
        ],
    )

    assert add_result

    mult_result = runner_pipeline_execute(
        runner,
        [
            '-y',
            file_relative_path(__file__, '../repository.yaml'),
            '--env',
            file_relative_path(
                __file__, '../environments/multi_mode_with_resources/mult_mode.yaml'
            ),
            '-d',
            'mult_mode',
            'multi_mode_with_resources',  # pipeline name
        ],
    )

    assert mult_result

    double_adder_result = runner_pipeline_execute(
        runner,
        [
            '-y',
            file_relative_path(__file__, '../repository.yaml'),
            '--env',
            file_relative_path(
                __file__, '../environments/multi_mode_with_resources/double_adder_mode.yaml'
            ),
            '-d',
            'double_adder_mode',
            'multi_mode_with_resources',  # pipeline name
        ],
    )

    assert double_adder_result


def test_execute_preset_command():
    runner = CliRunner()
    add_result = runner_pipeline_execute(
        runner,
        [
            '-y',
            file_relative_path(__file__, '../repository.yaml'),
            '-p',
            'add',
            'multi_mode_with_resources',  # pipeline name
        ],
    )

    assert 'PIPELINE_SUCCESS' in add_result.output

    # Can't use -p with --env
    bad_res = runner.invoke(
        pipeline_execute_command,
        [
            '-y',
            file_relative_path(__file__, '../repository.yaml'),
            '-p',
            'add',
            '--env',
            file_relative_path(
                __file__, '../environments/multi_mode_with_resources/double_adder_mode.yaml'
            ),
            'multi_mode_with_resources',  # pipeline name
        ],
    )
    assert bad_res.exit_code == 2


def test_execute_command():
    for cli_args in valid_execute_args():
        execute_execute_command(env=None, cli_args=cli_args)

    for cli_args in valid_execute_args():
        execute_execute_command(
            env=[file_relative_path(__file__, 'default_log_error_env.yaml')], cli_args=cli_args
        )

    runner = CliRunner()

    for cli_args in valid_cli_args():
        runner_pipeline_execute(runner, cli_args)

        runner_pipeline_execute(
            runner, ['--env', file_relative_path(__file__, 'default_log_error_env.yaml')] + cli_args
        )

    res = runner.invoke(
        pipeline_execute_command,
        [
            '-y',
            file_relative_path(__file__, 'repository_module.yaml'),
            'hello_cereal_pipeline',
            'foo',
        ],
    )
    assert res.exit_code == 1
    assert isinstance(res.exception, CheckError)
    assert 'Can only handle zero or one pipeline args.' in str(res.exception)


def test_stdout_execute_command():
    runner = CliRunner()
    result = runner_pipeline_execute(
        runner,
        ['-f', file_relative_path(__file__, 'test_cli_commands.py'), '-n', 'stdout_pipeline'],
    )
    assert 'HELLO WORLD' in result.output


def test_stderr_execute_command():
    runner = CliRunner()
    result = runner_pipeline_execute(
        runner,
        ['-f', file_relative_path(__file__, 'test_cli_commands.py'), '-n', 'stderr_pipeline'],
    )
    assert 'I AM SUPPOSED TO FAIL' in result.output


def test_fn_not_found_execute():
    with pytest.raises(DagsterInvariantViolationError) as exc_info:
        execute_execute_command(
            env=None,
            cli_args={
                'repository_yaml': None,
                'pipeline_name': (),
                'python_file': file_relative_path(__file__, 'test_cli_commands.py'),
                'module_name': None,
                'fn_name': 'nope',
            },
        )

    assert 'nope not found in module' in str(exc_info.value)


def not_a_repo_or_pipeline_fn():
    return 'kdjfkjdf'


not_a_repo_or_pipeline = 123


def test_fn_is_wrong_thing():
    with pytest.raises(DagsterInvariantViolationError) as exc_info:
        execute_execute_command(
            env={},
            cli_args={
                'repository_yaml': None,
                'pipeline_name': (),
                'python_file': file_relative_path(__file__, 'test_cli_commands.py'),
                'module_name': None,
                'fn_name': 'not_a_repo_or_pipeline',
            },
        )

    assert str(exc_info.value) == (
        'not_a_repo_or_pipeline must be a function that returns a '
        'PipelineDefinition or a RepositoryDefinition, or a function '
        'decorated with @pipeline.'
    )


def test_fn_returns_wrong_thing():
    with pytest.raises(DagsterInvariantViolationError) as exc_info:
        execute_execute_command(
            env={},
            cli_args={
                'repository_yaml': None,
                'pipeline_name': (),
                'python_file': file_relative_path(__file__, 'test_cli_commands.py'),
                'module_name': None,
                'fn_name': 'not_a_repo_or_pipeline_fn',
            },
        )

    assert str(exc_info.value) == (
        'not_a_repo_or_pipeline_fn is a function but must return a '
        'PipelineDefinition or a RepositoryDefinition, or be decorated '
        'with @pipeline.'
    )


def runner_pipeline_execute(runner, cli_args):
    result = runner.invoke(pipeline_execute_command, cli_args)
    if result.exit_code != 0:
        # CliRunner captures stdout so printing it out here
        raise Exception(
            (
                'dagster pipeline execute commands with cli_args {cli_args} '
                'returned exit_code {exit_code} with stdout:\n"{stdout}" and '
                '\nresult as string: "{result}"'
            ).format(
                cli_args=cli_args, exit_code=result.exit_code, stdout=result.stdout, result=result
            )
        )
    return result


def test_scaffold_command():
    for cli_args in valid_execute_args():
        cli_args['print_only_required'] = True
        execute_scaffold_command(cli_args=cli_args, print_fn=no_print)

        cli_args['print_only_required'] = False
        execute_scaffold_command(cli_args=cli_args, print_fn=no_print)

    runner = CliRunner()

    for cli_args in valid_cli_args():
        result = runner.invoke(pipeline_scaffold_command, cli_args)
        assert result.exit_code == 0

        result = runner.invoke(pipeline_scaffold_command, ['-p'] + cli_args)
        assert result.exit_code == 0


def test_default_memory_run_storage():
    cli_args = {
        'repository_yaml': file_relative_path(__file__, 'repository_file.yaml'),
        'pipeline_name': ('foo',),
        'python_file': None,
        'module_name': None,
        'fn_name': None,
    }
    result = execute_execute_command(env=None, cli_args=cli_args)
    assert result.success


def test_override_with_in_memory_storage():
    cli_args = {
        'repository_yaml': file_relative_path(__file__, 'repository_file.yaml'),
        'pipeline_name': ('foo',),
        'python_file': None,
        'module_name': None,
        'fn_name': None,
    }
    result = execute_execute_command(
        env=[file_relative_path(__file__, 'in_memory_env.yaml')], cli_args=cli_args
    )
    assert result.success


def test_override_with_filesystem_storage():
    cli_args = {
        'repository_yaml': file_relative_path(__file__, 'repository_file.yaml'),
        'pipeline_name': ('foo',),
        'python_file': None,
        'module_name': None,
        'fn_name': None,
    }
    result = execute_execute_command(
        env=[file_relative_path(__file__, 'filesystem_env.yaml')], cli_args=cli_args
    )
    assert result.success


def test_run_list():
    runner = CliRunner()
    result = runner.invoke(run_list_command)
    assert result.exit_code == 0


def test_run_wipe_correct_delete_message():
    runner = CliRunner()
    result = runner.invoke(run_wipe_command, input="DELETE\n")
    assert 'Deleted all run history and event logs' in result.output
    assert result.exit_code == 0


def test_run_wipe_incorrect_delete_message():
    runner = CliRunner()
    result = runner.invoke(run_wipe_command, input="WRONG\n")
    assert 'Exiting without deleting all run history and event logs' in result.output
    assert result.exit_code == 0


def define_bar_schedules():
    return [
        ScheduleDefinition(
            "foo_schedule",
            cron_schedule="* * * * *",
            pipeline_name="test_pipeline",
            environment_dict={},
        )
    ]


@pytest.fixture(name="scheduler_instance")
def define_scheduler_instance():
    with seven.TemporaryDirectory() as temp_dir:
        yield DagsterInstance(
            instance_type=InstanceType.EPHEMERAL,
            local_artifact_storage=LocalArtifactStorage(temp_dir),
            run_storage=InMemoryRunStorage(),
            event_storage=InMemoryEventLogStorage(),
            schedule_storage=SqliteScheduleStorage.from_local(temp_dir),
            scheduler=FilesytemTestScheduler(temp_dir),
            compute_log_manager=NoOpComputeLogManager(temp_dir),
        )


@pytest.fixture(name="_patch_scheduler_instance")
def mock_scheduler_instance(mocker, scheduler_instance):
    mocker.patch(
        'dagster.core.instance.DagsterInstance.get', return_value=scheduler_instance,
    )


def test_schedules_list(_patch_scheduler_instance):
    runner = CliRunner()

    result = runner.invoke(
        schedule_list_command, ['-y', file_relative_path(__file__, 'repository_file.yaml')]
    )

    if result.exception:
        raise result.exception

    assert result.exit_code == 0
    assert result.output == ('Repository bar\n' '**************\n')


def test_schedules_up(_patch_scheduler_instance):
    runner = CliRunner()

    result = runner.invoke(
        schedule_up_command, ['-y', file_relative_path(__file__, 'repository_file.yaml')]
    )

    assert result.exit_code == 0
    assert result.output == 'Changes:\n  + foo_schedule (add)\n'


def test_schedules_up_and_list(_patch_scheduler_instance):
    runner = CliRunner()

    result = runner.invoke(
        schedule_up_command, ['-y', file_relative_path(__file__, 'repository_file.yaml')]
    )

    result = runner.invoke(
        schedule_list_command, ['-y', file_relative_path(__file__, 'repository_file.yaml')]
    )

    assert result.exit_code == 0
    assert (
        result.output == 'Repository bar\n'
        '**************\n'
        'Schedule: foo_schedule [STOPPED]\n'
        'Cron Schedule: * * * * *\n'
    )


def test_schedules_start_and_stop(_patch_scheduler_instance):
    runner = CliRunner()

    result = runner.invoke(
        schedule_up_command, ['-y', file_relative_path(__file__, 'repository_file.yaml')],
    )

    result = runner.invoke(
        schedule_start_command,
        ['-y', file_relative_path(__file__, 'repository_file.yaml'), 'foo_schedule'],
    )

    assert result.exit_code == 0
    assert 'Started schedule foo_schedule\n' == result.output

    result = runner.invoke(
        schedule_stop_command,
        ['-y', file_relative_path(__file__, 'repository_file.yaml'), 'foo_schedule'],
    )

    assert result.exit_code == 0
    assert 'Stopped schedule foo_schedule\n' == result.output


def test_schedules_start_empty(_patch_scheduler_instance):
    runner = CliRunner()

    result = runner.invoke(
        schedule_start_command, ['-y', file_relative_path(__file__, 'repository_file.yaml')],
    )

    assert result.exit_code == 0
    assert 'Noop: dagster schedule start was called without any arguments' in result.output


def test_schedules_start_all(_patch_scheduler_instance):
    runner = CliRunner()

    result = runner.invoke(
        schedule_up_command, ['-y', file_relative_path(__file__, 'repository_file.yaml')]
    )

    result = runner.invoke(
        schedule_start_command,
        ['-y', file_relative_path(__file__, 'repository_file.yaml'), '--start-all'],
    )

    assert result.exit_code == 0
    assert result.output == 'Started all schedules for repository bar\n'


def test_schedules_wipe_correct_delete_message(_patch_scheduler_instance):
    runner = CliRunner()

    result = runner.invoke(
        schedule_up_command, ['-y', file_relative_path(__file__, 'repository_file.yaml')]
    )

    result = runner.invoke(
        schedule_wipe_command,
        ['-y', file_relative_path(__file__, 'repository_file.yaml')],
        input="DELETE\n",
    )

    if result.exception:
        raise result.exception

    assert result.exit_code == 0
    assert 'Wiped all schedules and schedule cron jobs' in result.output

    result = runner.invoke(
        schedule_up_command,
        ['-y', file_relative_path(__file__, 'repository_file.yaml'), '--preview'],
    )

    # Verify schedules were wiped
    assert result.exit_code == 0
    assert result.output == 'Planned Changes:\n  + foo_schedule (add)\n'


def test_schedules_wipe_incorrect_delete_message(_patch_scheduler_instance):
    runner = CliRunner()

    result = runner.invoke(
        schedule_up_command, ['-y', file_relative_path(__file__, 'repository_file.yaml')]
    )

    result = runner.invoke(
        schedule_wipe_command,
        ['-y', file_relative_path(__file__, 'repository_file.yaml')],
        input="WRONG\n",
    )

    assert result.exit_code == 0
    assert 'Exiting without deleting all schedules and schedule cron jobs' in result.output

    result = runner.invoke(
        schedule_up_command,
        ['-y', file_relative_path(__file__, 'repository_file.yaml'), '--preview'],
    )

    # Verify schedules were not wiped
    assert result.exit_code == 0
    assert result.output == 'No planned changes to schedules.\n1 schedules will remain unchanged\n'


def test_schedules_restart(_patch_scheduler_instance):
    runner = CliRunner()

    result = runner.invoke(
        schedule_up_command, ['-y', file_relative_path(__file__, 'repository_file.yaml')]
    )

    result = runner.invoke(
        schedule_start_command,
        ['-y', file_relative_path(__file__, 'repository_file.yaml'), 'foo_schedule'],
    )

    result = runner.invoke(
        schedule_restart_command,
        ['-y', file_relative_path(__file__, 'repository_file.yaml'), 'foo_schedule'],
    )

    assert result.exit_code == 0
    assert 'Restarted schedule foo_schedule' in result.output


def test_schedules_restart_all(_patch_scheduler_instance):
    runner = CliRunner()

    result = runner.invoke(
        schedule_up_command, ['-y', file_relative_path(__file__, 'repository_file.yaml')]
    )

    result = runner.invoke(
        schedule_start_command,
        ['-y', file_relative_path(__file__, 'repository_file.yaml'), 'foo_schedule'],
    )

    result = runner.invoke(
        schedule_restart_command,
        [
            '-y',
            file_relative_path(__file__, 'repository_file.yaml'),
            'foo_schedule',
            '--restart-all-running',
        ],
    )

    assert result.exit_code == 0
    assert result.output == 'Restarted all running schedules for repository bar\n'


@pytest.mark.skipif(
    os.name == 'nt', reason="multiproc directory test disabled for windows because of fs contention"
)
def test_multiproc():
    with seven.TemporaryDirectory() as temp:
        runner = CliRunner(env={'DAGSTER_HOME': temp})
        add_result = runner_pipeline_execute(
            runner,
            [
                '-y',
                file_relative_path(__file__, '../repository.yaml'),
                '-p',
                'multiproc',
                'multi_mode_with_resources',  # pipeline name
            ],
        )
        assert 'PIPELINE_SUCCESS' in add_result.output


def test_multiproc_invalid():
    # force ephemeral instance by removing out DAGSTER_HOME
    runner = CliRunner(env={'DAGSTER_HOME': None})
    add_result = runner_pipeline_execute(
        runner,
        [
            '-y',
            file_relative_path(__file__, '../repository.yaml'),
            '-p',
            'multiproc',
            'multi_mode_with_resources',  # pipeline name
        ],
    )
    # which is invalid for multiproc
    assert 'DagsterUnmetExecutorRequirementsError' in add_result.output


class InMemoryRunLauncher(RunLauncher, ConfigurableClass):
    def __init__(self, inst_data=None):
        self._inst_data = inst_data
        self._queue = []

    def launch_run(self, instance, run):
        self._queue.append(run)
        return run

    def queue(self):
        return self._queue

    @classmethod
    def config_type(cls):
        return Shape({})

    @classmethod
    def from_config_value(cls, inst_data, config_value):
        return cls(inst_data=inst_data,)

    @property
    def inst_data(self):
        return self._inst_data


def define_baz_partitions():
    return [
        PartitionSetDefinition(
            name='baz_partitions',
            pipeline_name='baz',
            partition_fn=lambda: string.ascii_lowercase,
            environment_dict_fn_for_partition=lambda partition: {
                'solids': {'do_input': {'inputs': {'x': {'value': partition}}}}
            },
        )
    ]


def backfill_execute_args(execution_args):
    backfill_args = {
        'repository_yaml': file_relative_path(__file__, 'repository_file.yaml'),
        'noprompt': True,
    }
    pipeline_name = execution_args.get('pipeline_name')
    if pipeline_name:
        backfill_args['pipeline_name'] = (pipeline_name,)
    for name, value in execution_args.items():
        if name != 'pipeline_name':
            backfill_args[name] = value
    return backfill_args


def backfill_cli_runner_args(execution_args):
    backfill_args = ['-y', file_relative_path(__file__, 'repository_file.yaml'), '--noprompt']
    pipeline_name = execution_args.get('pipeline_name')
    if pipeline_name:
        backfill_args.append(pipeline_name)
    for name, value in execution_args.items():
        if name != 'pipeline_name':
            backfill_args.extend(['--{}'.format(name.replace('_', '-')), value])
    return backfill_args


def run_test_backfill(
    execution_args, expected_count=None, error_message=None, use_run_launcher=True
):
    runner = CliRunner()
    run_launcher = InMemoryRunLauncher() if use_run_launcher else None
    with seven.TemporaryDirectory() as temp_dir:
        instance = DagsterInstance(
            instance_type=InstanceType.EPHEMERAL,
            local_artifact_storage=LocalArtifactStorage(temp_dir),
            run_storage=InMemoryRunStorage(),
            event_storage=InMemoryEventLogStorage(),
            compute_log_manager=NoOpComputeLogManager(temp_dir),
            run_launcher=run_launcher,
        )
        with mock.patch('dagster.core.instance.DagsterInstance.get') as _instance:
            _instance.return_value = instance

            if error_message:
                with pytest.raises(UsageError) as error_info:
                    execute_backfill_command(backfill_execute_args(execution_args), no_print)
                assert error_info and error_message in error_info.value.message

            result = runner.invoke(
                pipeline_backfill_command, backfill_cli_runner_args(execution_args)
            )
            if error_message:
                assert result.exit_code == 2
            else:
                assert result.exit_code == 0
                if expected_count:
                    assert len(run_launcher.queue()) == expected_count


def test_backfill_no_run_launcher():
    args = {'pipeline_name': 'baz'}  # legit partition args
    run_test_backfill(
        args, use_run_launcher=False, error_message='A run launcher must be configured'
    )


def test_backfill_no_pipeline():
    args = {'pipeline_name': 'nonexistent'}
    run_test_backfill(args, error_message='No pipeline found')


def test_backfill_no_partition_sets():
    args = {'pipeline_name': 'foo'}
    run_test_backfill(args, error_message='No partition sets found')


def test_backfill_no_named_partition_set():
    args = {'pipeline_name': 'baz', 'partition_set': 'nonexistent'}
    run_test_backfill(args, error_message='No partition set found')


def test_backfill_launch():
    args = {'pipeline_name': 'baz', 'partition_set': 'baz_partitions'}
    run_test_backfill(args, expected_count=len(string.ascii_lowercase))


def test_backfill_partition_range():
    args = {'pipeline_name': 'baz', 'partition_set': 'baz_partitions', 'from': 'x'}
    run_test_backfill(args, expected_count=3)

    args = {'pipeline_name': 'baz', 'partition_set': 'baz_partitions', 'to': 'c'}
    run_test_backfill(args, expected_count=3)

    args = {'pipeline_name': 'baz', 'partition_set': 'baz_partitions', 'from': 'c', 'to': 'f'}
    run_test_backfill(args, expected_count=4)


def test_backfill_partition_enum():
    args = {'pipeline_name': 'baz', 'partition_set': 'baz_partitions', 'partitions': 'c,x,z'}
    run_test_backfill(args, expected_count=3)


def run_launch(execution_args, expected_count=None):
    runner = CliRunner()
    run_launcher = InMemoryRunLauncher()
    with seven.TemporaryDirectory() as temp_dir:
        instance = DagsterInstance(
            instance_type=InstanceType.EPHEMERAL,
            local_artifact_storage=LocalArtifactStorage(temp_dir),
            run_storage=InMemoryRunStorage(),
            event_storage=InMemoryEventLogStorage(),
            compute_log_manager=NoOpComputeLogManager(temp_dir),
            run_launcher=run_launcher,
        )
        with mock.patch('dagster.core.instance.DagsterInstance.get') as _instance:
            _instance.return_value = instance

            result = runner.invoke(pipeline_launch_command, execution_args)
            assert result.exit_code == 0
            if expected_count:
                assert len(run_launcher.queue()) == expected_count


def test_launch_pipeline():
    for cli_args in valid_cli_args():
        run_launch(cli_args, expected_count=1)


@contextmanager
def mocked_instance():
    with seven.TemporaryDirectory() as temp_dir:
        instance = DagsterInstance(
            instance_type=InstanceType.EPHEMERAL,
            local_artifact_storage=LocalArtifactStorage(temp_dir),
            run_storage=InMemoryRunStorage(),
            event_storage=InMemoryEventLogStorage(),
            compute_log_manager=NoOpComputeLogManager(temp_dir),
            run_launcher=InMemoryRunLauncher(),
        )
        with mock.patch('dagster.core.instance.DagsterInstance.get') as _instance:
            _instance.return_value = instance
            yield instance


def test_tags_pipeline():
    runner = CliRunner()
    with mocked_instance() as instance:
        result = runner.invoke(
            pipeline_execute_command,
            [
                '-y',
                file_relative_path(__file__, 'repository_module.yaml'),
                '--tags',
                '{ "foo": "bar" }',
                'hello_cereal_pipeline',
            ],
        )
        assert result.exit_code == 0
        runs = instance.get_runs()
        assert len(runs) == 1
        run = runs[0]
        assert len(run.tags) == 1
        assert run.tags.get('foo') == 'bar'

    with mocked_instance() as instance:
        result = runner.invoke(
            pipeline_execute_command,
            [
                '-y',
                file_relative_path(__file__, '../repository.yaml'),
                '-p',
                'add',
                '--tags',
                '{ "foo": "bar" }',
                'multi_mode_with_resources',  # pipeline name
            ],
        )
        assert result.exit_code == 0
        runs = instance.get_runs()
        assert len(runs) == 1
        run = runs[0]
        assert len(run.tags) == 1
        assert run.tags.get('foo') == 'bar'

    with mocked_instance() as instance:
        result = runner.invoke(
            pipeline_backfill_command,
            [
                '-y',
                file_relative_path(__file__, 'repository_file.yaml'),
                '--noprompt',
                '--partition-set',
                'baz_partitions',
                '--partitions',
                'c',
                '--tags',
                '{ "foo": "bar" }',
                'baz',
            ],
        )
        assert result.exit_code == 0
        runs = instance.run_launcher.queue()
        assert len(runs) == 1
        run = runs[0]
        assert len(run.tags) >= 1
        assert run.tags.get('foo') == 'bar'



from abc import ABCMeta, abstractmethod

import six


class ExecutablePipeline(six.with_metaclass(ABCMeta)):
    @abstractmethod
    def get_definition(self):
        pass

    @abstractmethod
    def build_sub_pipeline(self, solid_subset):
        pass


class InMemoryExecutablePipeline(ExecutablePipeline, object):
    def __init__(self, pipeline_def):
        self._pipeline_def = pipeline_def

    def get_definition(self):
        return self._pipeline_def

    def build_sub_pipeline(self, solid_subset):
        return InMemoryExecutablePipeline(self._pipeline_def.build_sub_pipeline(solid_subset))



import os
import warnings
from uuid import uuid4

from dagster import DagsterInvariantViolationError, check
from dagster.core.snap import ActiveRepositoryData
from dagster.serdes import deserialize_json_to_dagster_namedtuple
from dagster.seven import get_system_temp_directory
from dagster.utils.temp_file import get_temp_dir

DEFAULT_INTERNAL_VOLUME = '/data'
DEFAULT_MODE = 'rw'


def run_serialized_container_command(image, command, volumes):
    try:
        from docker.client import from_env

        client = from_env()
        client.containers.run(
            image, command=command, detach=False, volumes=volumes, auto_remove=True,
        )
    except ImportError:
        warnings.warn(
            "Cannot load docker environment without the python package docker. Ensure that dagster[docker] or the python package docker is installed."
        )
        raise


def _get_active_repo_data(path):
    with open(path, 'r') as fp:
        return deserialize_json_to_dagster_namedtuple(fp.read())


def get_active_repository_data_from_image(image):
    check.str_param(image, 'image')

    with get_temp_dir(in_directory=get_system_temp_directory()) as tmp_dir:
        output_file_name = "{}.json".format(uuid4())
        run_serialized_container_command(
            image=image,
            command='dagster repository snapshot {output_file}'.format(
                output_file=os.path.join(DEFAULT_INTERNAL_VOLUME, output_file_name)
            ),
            volumes={tmp_dir: {'bind': DEFAULT_INTERNAL_VOLUME, 'mode': DEFAULT_MODE}},
        )

        active_repo_data = _get_active_repo_data(os.path.join(tmp_dir, output_file_name))
        if not isinstance(active_repo_data, ActiveRepositoryData):
            raise DagsterInvariantViolationError(
                "Deserialized snapshot is of type {received} must be a ActiveRepositoryData".format(
                    received=type(active_repo_data)
                )
            )
        return active_repo_data



import sys
from collections import defaultdict
from contextlib import contextmanager

import pytest

from dagster import check
from dagster.check import (
    CheckError,
    ElementCheckError,
    NotImplementedCheckError,
    ParameterCheckError,
)
from dagster.utils import frozendict, frozenlist


def test_int_param():
    assert check.int_param(-1, 'param_name') == -1
    assert check.int_param(0, 'param_name') == 0
    assert check.int_param(1, 'param_name') == 1

    with pytest.raises(ParameterCheckError):
        check.int_param(None, 'param_name')

    with pytest.raises(ParameterCheckError):
        check.int_param('s', 'param_name')


def test_int_value_param():
    assert check.int_value_param(-1, -1, 'param_name') == -1
    with pytest.raises(ParameterCheckError):
        check.int_value_param(None, -1, 'param_name')

    with pytest.raises(ParameterCheckError):
        check.int_value_param(1, 0, 'param_name')


def test_opt_int_param():
    assert check.opt_int_param(-1, 'param_name') == -1
    assert check.opt_int_param(0, 'param_name') == 0
    assert check.opt_int_param(1, 'param_name') == 1
    assert check.opt_int_param(None, 'param_name') is None

    with pytest.raises(ParameterCheckError):
        check.opt_int_param('s', 'param_name')


def test_float_param():
    assert check.float_param(-1.0, 'param_name') == -1.0
    assert check.float_param(0.0, 'param_name') == 0.0
    assert check.float_param(1.1, 'param_name') == 1.1

    with pytest.raises(ParameterCheckError):
        check.float_param(None, 'param_name')

    with pytest.raises(ParameterCheckError):
        check.float_param('s', 'param_name')

    with pytest.raises(ParameterCheckError):
        check.float_param(1, 'param_name')

    with pytest.raises(ParameterCheckError):
        check.float_param(0, 'param_name')


def test_opt_float_param():
    assert check.opt_float_param(-1.0, 'param_name') == -1.0
    assert check.opt_float_param(0.0, 'param_name') == 0.0
    assert check.opt_float_param(1.1, 'param_name') == 1.1
    assert check.opt_float_param(None, 'param_name') is None

    with pytest.raises(ParameterCheckError):
        check.opt_float_param('s', 'param_name')


def test_list_param():
    assert check.list_param([], 'list_param') == []
    assert check.list_param(frozenlist(), 'list_param') == []

    with pytest.raises(ParameterCheckError):
        check.list_param(None, 'list_param')

    with pytest.raises(ParameterCheckError):
        check.list_param('3u4', 'list_param')


def test_set_param():
    assert check.set_param(set(), 'set_param') == set()
    assert check.set_param(frozenset(), 'set_param') == set()

    with pytest.raises(ParameterCheckError):
        check.set_param(None, 'set_param')

    with pytest.raises(ParameterCheckError):
        check.set_param('3u4', 'set_param')

    obj_set = {1}
    assert check.set_param(obj_set, 'set_param') == obj_set

    obj_set_two = {1, 1, 2}
    obj_set_two_deduped = {1, 2}
    assert check.set_param(obj_set_two, 'set_param') == obj_set_two_deduped
    assert check.set_param(obj_set_two, 'set_param', of_type=int) == obj_set_two_deduped

    with pytest.raises(CheckError, match='Did you pass a class'):
        check.set_param({str}, 'set_param', of_type=int)

    with pytest.raises(CheckError, match='Member of set mismatches type'):
        check.set_param({'foo'}, 'set_param', of_type=int)


def test_is_list():
    assert check.is_list([]) == []

    with pytest.raises(CheckError):
        check.is_list(None)

    with pytest.raises(CheckError):
        check.is_list('3u4')

    with pytest.raises(CheckError, match='Did you pass a class'):
        check.is_list([str], of_type=int)


def test_typed_list_param():
    class Foo(object):
        pass

    class Bar(object):
        pass

    assert check.list_param([], 'list_param', Foo) == []
    foo_list = [Foo()]
    assert check.list_param(foo_list, 'list_param', Foo) == foo_list

    with pytest.raises(CheckError):
        check.list_param([Bar()], 'list_param', Foo)

    with pytest.raises(CheckError):
        check.list_param([None], 'list_param', Foo)


def test_typed_is_list():
    class Foo(object):
        pass

    class Bar(object):
        pass

    assert check.is_list([], Foo) == []
    foo_list = [Foo()]
    assert check.is_list(foo_list, Foo) == foo_list

    with pytest.raises(CheckError):
        check.is_list([Bar()], Foo)

    with pytest.raises(CheckError):
        check.is_list([None], Foo)


def test_opt_list_param():
    assert check.opt_list_param(None, 'list_param') == []
    assert check.opt_list_param([], 'list_param') == []
    assert check.opt_list_param(frozenlist(), 'list_param') == []
    obj_list = [1]
    assert check.list_param(obj_list, 'list_param') == obj_list

    with pytest.raises(ParameterCheckError):
        check.opt_list_param(0, 'list_param')

    with pytest.raises(ParameterCheckError):
        check.opt_list_param('', 'list_param')

    with pytest.raises(ParameterCheckError):
        check.opt_list_param('3u4', 'list_param')


def test_opt_set_param():
    assert check.opt_set_param(None, 'set_param') == set()
    assert check.opt_set_param(set(), 'set_param') == set()
    assert check.opt_set_param(frozenset(), 'set_param') == set()
    assert check.opt_set_param({3}, 'set_param') == {3}

    with pytest.raises(ParameterCheckError):
        check.opt_set_param(0, 'set_param')

    with pytest.raises(ParameterCheckError):
        check.opt_set_param('', 'set_param')

    with pytest.raises(ParameterCheckError):
        check.opt_set_param('3u4', 'set_param')


def test_opt_nullable_list_param():
    assert check.opt_nullable_list_param(None, 'list_param') is None
    assert check.opt_nullable_list_param([], 'list_param') == []
    assert check.opt_nullable_list_param(frozenlist(), 'list_param') == []
    obj_list = [1]
    assert check.opt_nullable_list_param(obj_list, 'list_param') == obj_list

    with pytest.raises(ParameterCheckError):
        check.opt_nullable_list_param(0, 'list_param')

    with pytest.raises(ParameterCheckError):
        check.opt_nullable_list_param('', 'list_param')

    with pytest.raises(ParameterCheckError):
        check.opt_nullable_list_param('3u4', 'list_param')


def test_opt_typed_list_param():
    class Foo(object):
        pass

    class Bar(object):
        pass

    assert check.opt_list_param(None, 'list_param', Foo) == []
    assert check.opt_list_param([], 'list_param', Foo) == []
    foo_list = [Foo()]
    assert check.opt_list_param(foo_list, 'list_param', Foo) == foo_list

    with pytest.raises(CheckError):
        check.opt_list_param([Bar()], 'list_param', Foo)

    with pytest.raises(CheckError):
        check.opt_list_param([None], 'list_param', Foo)


def test_dict_param():
    assert check.dict_param({}, 'dict_param') == {}
    assert check.dict_param(frozendict(), 'dict_param') == {}
    ddict = {'a': 2}
    assert check.dict_param(ddict, 'dict_param') == ddict

    with pytest.raises(ParameterCheckError):
        check.dict_param(None, 'dict_param')

    with pytest.raises(ParameterCheckError):
        check.dict_param(0, 'dict_param')

    with pytest.raises(ParameterCheckError):
        check.dict_param(1, 'dict_param')

    with pytest.raises(ParameterCheckError):
        check.dict_param('foo', 'dict_param')

    with pytest.raises(ParameterCheckError):
        check.dict_param(['foo'], 'dict_param')

    with pytest.raises(ParameterCheckError):
        check.dict_param([], 'dict_param')


def test_dict_param_with_type():
    str_to_int = {'str': 1}
    assert check.dict_param(str_to_int, 'str_to_int', key_type=str, value_type=int)
    assert check.dict_param(str_to_int, 'str_to_int', value_type=int)
    assert check.dict_param(str_to_int, 'str_to_int', key_type=str)
    assert check.dict_param(str_to_int, 'str_to_int')

    assert check.dict_param({}, 'str_to_int', key_type=str, value_type=int) == {}
    assert check.dict_param({}, 'str_to_int', value_type=int) == {}
    assert check.dict_param({}, 'str_to_int', key_type=str) == {}
    assert check.dict_param({}, 'str_to_int') == {}

    class Wrong(object):
        pass

    with pytest.raises(CheckError):
        assert check.dict_param(str_to_int, 'str_to_int', key_type=Wrong, value_type=Wrong)

    with pytest.raises(CheckError):
        assert check.dict_param(str_to_int, 'str_to_int', key_type=Wrong, value_type=int)

    with pytest.raises(CheckError):
        assert check.dict_param(str_to_int, 'str_to_int', key_type=str, value_type=Wrong)

    with pytest.raises(CheckError):
        assert check.dict_param(str_to_int, 'str_to_int', key_type=Wrong)

    with pytest.raises(CheckError):
        assert check.dict_param(str_to_int, 'str_to_int', value_type=Wrong)


def test_opt_dict_param_with_type():
    str_to_int = {'str': 1}
    assert check.opt_dict_param(str_to_int, 'str_to_int', key_type=str, value_type=int)
    assert check.opt_dict_param(str_to_int, 'str_to_int', value_type=int)
    assert check.opt_dict_param(str_to_int, 'str_to_int', key_type=str)
    assert check.opt_dict_param(str_to_int, 'str_to_int')

    assert check.opt_dict_param({}, 'str_to_int', key_type=str, value_type=int) == {}
    assert check.opt_dict_param({}, 'str_to_int', value_type=int) == {}
    assert check.opt_dict_param({}, 'str_to_int', key_type=str) == {}
    assert check.opt_dict_param({}, 'str_to_int') == {}

    assert check.opt_dict_param(None, 'str_to_int', key_type=str, value_type=int) == {}
    assert check.opt_dict_param(None, 'str_to_int', value_type=int) == {}
    assert check.opt_dict_param(None, 'str_to_int', key_type=str) == {}
    assert check.opt_dict_param(None, 'str_to_int') == {}

    class Wrong(object):
        pass

    with pytest.raises(CheckError):
        assert check.opt_dict_param(str_to_int, 'str_to_int', key_type=Wrong, value_type=Wrong)

    with pytest.raises(CheckError):
        assert check.opt_dict_param(str_to_int, 'str_to_int', key_type=Wrong, value_type=int)

    with pytest.raises(CheckError):
        assert check.opt_dict_param(str_to_int, 'str_to_int', key_type=str, value_type=Wrong)

    with pytest.raises(CheckError):
        assert check.opt_dict_param(str_to_int, 'str_to_int', key_type=Wrong)

    with pytest.raises(CheckError):
        assert check.opt_dict_param(str_to_int, 'str_to_int', value_type=Wrong)


def test_opt_dict_param():
    assert check.opt_dict_param(None, 'opt_dict_param') == {}
    assert check.opt_dict_param({}, 'opt_dict_param') == {}
    assert check.opt_dict_param(frozendict(), 'opt_dict_param') == {}
    ddict = {'a': 2}
    assert check.opt_dict_param(ddict, 'opt_dict_param') == ddict

    with pytest.raises(ParameterCheckError):
        check.opt_dict_param(0, 'opt_dict_param')

    with pytest.raises(ParameterCheckError):
        check.opt_dict_param(1, 'opt_dict_param')

    with pytest.raises(ParameterCheckError):
        check.opt_dict_param('foo', 'opt_dict_param')

    with pytest.raises(ParameterCheckError):
        check.opt_dict_param(['foo'], 'opt_dict_param')

    with pytest.raises(ParameterCheckError):
        check.opt_dict_param([], 'opt_dict_param')


def test_opt_nullable_dict_param():
    assert check.opt_nullable_dict_param(None, 'opt_nullable_dict_param') is None
    assert check.opt_nullable_dict_param({}, 'opt_nullable_dict_param') == {}
    assert check.opt_nullable_dict_param(frozendict(), 'opt_nullable_dict_param') == {}
    ddict = {'a': 2}
    assert check.opt_nullable_dict_param(ddict, 'opt_nullable_dict_param') == ddict

    class Foo(object):
        pass

    class Bar(Foo):
        pass

    ddict_class = {'a': Bar}
    assert (
        check.opt_nullable_dict_param(ddict_class, 'opt_nullable_dict_param', value_class=Foo)
        == ddict_class
    )

    with pytest.raises(ParameterCheckError):
        check.opt_nullable_dict_param(1, 'opt_nullable_dict_param')

    with pytest.raises(ParameterCheckError):
        check.opt_nullable_dict_param('foo', 'opt_nullable_dict_param')


def test_str_param():
    assert check.str_param('a', 'str_param') == 'a'
    assert check.str_param('', 'str_param') == ''
    assert check.str_param(u'a', 'unicode_param') == u'a'

    with pytest.raises(ParameterCheckError):
        check.str_param(None, 'str_param')

    with pytest.raises(ParameterCheckError):
        check.str_param(0, 'str_param')

    with pytest.raises(ParameterCheckError):
        check.str_param(1, 'str_param')


def test_opt_str_param():
    assert check.opt_str_param('a', 'str_param') == 'a'
    assert check.opt_str_param('', 'str_param') == ''
    assert check.opt_str_param(u'a', 'unicode_param') == u'a'
    assert check.opt_str_param(None, 'str_param') is None
    assert check.opt_str_param(None, 'str_param', 'foo') == 'foo'

    with pytest.raises(ParameterCheckError):
        check.opt_str_param(0, 'str_param')

    with pytest.raises(ParameterCheckError):
        check.opt_str_param(1, 'str_param')


def test_opt_nonempty_str_param():
    assert check.opt_nonempty_str_param('a', 'str_param') == 'a'
    assert check.opt_nonempty_str_param('', 'str_param') is None
    assert check.opt_nonempty_str_param('', 'str_param', 'foo') == 'foo'
    assert check.opt_nonempty_str_param(u'a', 'unicode_param') == u'a'
    assert check.opt_nonempty_str_param(None, 'str_param') is None
    assert check.opt_nonempty_str_param(None, 'str_param', 'foo') == 'foo'

    with pytest.raises(ParameterCheckError):
        check.opt_nonempty_str_param(0, 'str_param')

    with pytest.raises(ParameterCheckError):
        check.opt_nonempty_str_param(1, 'str_param')


def test_bool_param():
    assert check.bool_param(True, 'b') is True
    assert check.bool_param(False, 'b') is False

    with pytest.raises(ParameterCheckError):
        check.bool_param(None, 'b')

    with pytest.raises(ParameterCheckError):
        check.bool_param(0, 'b')

    with pytest.raises(ParameterCheckError):
        check.bool_param('val', 'b')


def test_opt_bool_param():
    assert check.opt_bool_param(True, 'b') is True
    assert check.opt_bool_param(False, 'b') is False
    assert check.opt_bool_param(None, 'b') is None
    assert check.opt_bool_param(None, 'b', True) is True
    assert check.opt_bool_param(None, 'b', False) is False

    with pytest.raises(ParameterCheckError):
        check.opt_bool_param(0, 'b')

    with pytest.raises(ParameterCheckError):
        check.opt_bool_param('val', 'b')


def test_callable_param():
    lamb = lambda: 1
    assert check.callable_param(lamb, 'lamb') == lamb

    with pytest.raises(ParameterCheckError):
        check.callable_param(None, 'lamb')

    with pytest.raises(ParameterCheckError):
        check.callable_param(2, 'lamb')


def test_opt_callable_param():
    lamb = lambda: 1
    assert check.opt_callable_param(lamb, 'lamb') == lamb
    assert check.opt_callable_param(None, 'lamb') is None
    assert check.opt_callable_param(None, 'lamb', default=None) is None
    assert check.opt_callable_param(None, 'lamb', default=lamb) == lamb

    with pytest.raises(ParameterCheckError):
        check.opt_callable_param(2, 'lamb')


def test_param_invariant():
    check.param_invariant(True, 'some_param')
    num_to_check = 1
    check.param_invariant(num_to_check == 1, 'some_param')

    with pytest.raises(ParameterCheckError):
        check.param_invariant(num_to_check == 2, 'some_param')

    with pytest.raises(ParameterCheckError):
        check.param_invariant(False, 'some_param')

    with pytest.raises(ParameterCheckError):
        check.param_invariant(0, 'some_param')

    check.param_invariant(1, 'some_param')

    with pytest.raises(ParameterCheckError):
        check.param_invariant('', 'some_param')

    check.param_invariant('1kjkjsf', 'some_param')

    with pytest.raises(ParameterCheckError):
        check.param_invariant({}, 'some_param')

    check.param_invariant({234: '1kjkjsf'}, 'some_param')

    with pytest.raises(ParameterCheckError):
        check.param_invariant([], 'some_param')

    check.param_invariant([234], 'some_param')


def test_string_elem():
    ddict = {'a_str': 'a', 'a_num': 1, 'a_none': None}

    assert check.str_elem(ddict, 'a_str') == 'a'

    with pytest.raises(ElementCheckError):
        assert check.str_elem(ddict, 'a_none')

    with pytest.raises(ElementCheckError):
        check.str_elem(ddict, 'a_num')


def test_opt_string_elem():
    ddict = {'a_str': 'a', 'a_num': 1, 'a_none': None}

    assert check.opt_str_elem(ddict, 'a_str') == 'a'

    assert check.opt_str_elem(ddict, 'a_none') == None

    assert check.opt_str_elem(ddict, 'nonexistentkey') == None

    with pytest.raises(ElementCheckError):
        check.opt_str_elem(ddict, 'a_num')


def test_bool_elem():
    ddict = {'a_true': True, 'a_str': 'a', 'a_num': 1, 'a_none': None}

    assert check.bool_elem(ddict, 'a_true') is True

    with pytest.raises(ElementCheckError):
        check.bool_elem(ddict, 'a_none')

    with pytest.raises(ElementCheckError):
        check.bool_elem(ddict, 'a_num')

    with pytest.raises(ElementCheckError):
        check.bool_elem(ddict, 'a_str')


def test_invariant():
    assert check.invariant(True)

    with pytest.raises(CheckError):
        check.invariant(False)

    with pytest.raises(CheckError, match='Some Unique String'):
        check.invariant(False, 'Some Unique String')

    empty_list = []

    with pytest.raises(CheckError, match='Invariant failed'):
        check.invariant(empty_list)


def test_failed():
    with pytest.raises(CheckError, match='some desc'):
        check.failed('some desc')

    with pytest.raises(CheckError, match='must be a string'):
        check.failed(0)


def test_not_implemented():
    with pytest.raises(NotImplementedCheckError, match='some string'):
        check.not_implemented('some string')

    with pytest.raises(CheckError, match='desc argument must be a string'):
        check.not_implemented(None)


def test_inst():
    class Foo(object):
        pass

    class Bar(object):
        pass

    obj = Foo()

    assert check.inst(obj, Foo) == obj

    with pytest.raises(CheckError, match='not a Bar'):
        check.inst(Foo(), Bar)

    with pytest.raises(CheckError, match='Desc: Expected only a Bar'):
        check.inst(Foo(), Bar, 'Expected only a Bar')


def test_inst_param():
    class Foo(object):
        pass

    class Bar(object):
        pass

    class Baaz(object):
        pass

    obj = Foo()

    assert check.inst_param(obj, 'obj', Foo) == obj

    with pytest.raises(ParameterCheckError, match='not a Bar'):
        check.inst_param(None, 'obj', Bar)

    with pytest.raises(ParameterCheckError, match='not a Bar'):
        check.inst_param(Bar, 'obj', Bar)

    with pytest.raises(ParameterCheckError, match='not a Bar'):
        check.inst_param(Foo(), 'obj', Bar)

    with pytest.raises(ParameterCheckError, match=r"not one of \['Bar', 'Foo'\]"):
        check.inst_param(None, 'obj', (Foo, Bar))

    with pytest.raises(ParameterCheckError, match=r"not one of \['Bar', 'Foo'\]"):
        check.inst_param(Baaz(), 'obj', (Foo, Bar))


def test_opt_inst_param():
    class Foo(object):
        pass

    class Bar(object):
        pass

    class Baaz(object):
        pass

    obj = Foo()

    assert check.opt_inst_param(obj, 'obj', Foo) == obj
    assert check.opt_inst_param(None, 'obj', Foo) is None
    assert check.opt_inst_param(None, 'obj', Bar) is None

    with pytest.raises(ParameterCheckError, match='not a Bar'):
        check.opt_inst_param(Bar, 'obj', Bar)

    with pytest.raises(ParameterCheckError, match='not a Bar'):
        check.opt_inst_param(Foo(), 'obj', Bar)

    # check defaults

    default_obj = Foo()

    assert check.opt_inst_param(None, 'obj', Foo, default_obj) is default_obj

    assert check.opt_inst_param(None, 'obj', (Foo, Bar)) is None

    with pytest.raises(ParameterCheckError, match=r"not one of \['Bar', 'Foo'\]"):
        check.inst_param(Baaz(), 'obj', (Foo, Bar))


def test_dict_elem():
    dict_value = {'blah': 'blahblah'}
    ddict = {'dictkey': dict_value, 'stringkey': 'A', 'nonekey': None}

    assert check.dict_elem(ddict, 'dictkey') == dict_value

    with pytest.raises(CheckError):
        check.dict_elem(ddict, 'stringkey')

    with pytest.raises(CheckError):
        check.dict_elem(ddict, 'nonekey')

    with pytest.raises(CheckError):
        check.dict_elem(ddict, 'nonexistantkey')


def test_opt_dict_elem():
    dict_value = {'blah': 'blahblah'}
    ddict = {'dictkey': dict_value, 'stringkey': 'A', 'nonekey': None}

    assert check.opt_dict_elem(ddict, 'dictkey') == dict_value
    assert check.opt_dict_elem(ddict, 'nonekey') == {}
    assert check.opt_dict_elem(ddict, 'nonexistantkey') == {}

    with pytest.raises(CheckError):
        check.opt_dict_elem(ddict, 'stringkey')


def test_list_elem():
    list_value = ['blah', 'blahblah']
    ddict = {'listkey': list_value, 'stringkey': 'A', 'nonekey': None}

    assert check.list_elem(ddict, 'listkey') == list_value

    with pytest.raises(CheckError):
        assert check.list_elem(ddict, 'nonekey') == []

    with pytest.raises(CheckError):
        assert check.list_elem(ddict, 'nonexistantkey') == []

    with pytest.raises(CheckError):
        check.list_elem(ddict, 'stringkey')


def test_opt_list_elem():
    list_value = ['blah', 'blahblah']
    ddict = {'listkey': list_value, 'stringkey': 'A', 'nonekey': None}

    assert check.opt_list_elem(ddict, 'listkey') == list_value
    assert check.opt_list_elem(ddict, 'nonekey') == []
    assert check.opt_list_elem(ddict, 'nonexistantkey') == []

    with pytest.raises(CheckError):
        check.opt_list_elem(ddict, 'stringkey')


def test_not_none_param():
    assert check.not_none_param(1, 'fine')
    check.not_none_param(0, 'zero is fine')
    check.not_none_param('', 'empty str is fine')

    with pytest.raises(CheckError):
        check.not_none_param(None, 'none fails')


def test_is_callable():
    def fn():
        pass

    assert check.is_callable(fn) == fn
    assert check.is_callable(lambda: None)
    assert check.is_callable(lambda: None, 'some desc')

    with pytest.raises(CheckError):
        check.is_callable(None)

    with pytest.raises(CheckError):
        check.is_callable(1)

    with pytest.raises(CheckError, match='some other desc'):
        check.is_callable(1, 'some other desc')


def test_tuple_param():
    assert check.tuple_param((1, 2), 'something')

    with pytest.raises(CheckError):
        assert check.tuple_param(None, 'something')

    with pytest.raises(CheckError):
        assert check.tuple_param(1, 'something')

    with pytest.raises(CheckError):
        assert check.tuple_param([1], 'something')

    with pytest.raises(CheckError):
        assert check.tuple_param({1: 2}, 'something')

    with pytest.raises(CheckError):
        assert check.tuple_param('kdjfkd', 'something')

    assert check.tuple_param((3, 4), 'something', of_type=int)
    assert check.tuple_param(('foo', 'bar'), 'something', of_type=str)

    assert check.tuple_param((3, 4), 'something', of_type=(int, int))
    assert check.tuple_param((3, 4), 'something', of_type=(int, int))
    assert check.tuple_param((3, 'bar'), 'something', of_type=(int, str))

    with pytest.raises(CheckError):
        check.tuple_param((3, 4, 5), 'something', of_type=(int, int))

    with pytest.raises(CheckError):
        check.tuple_param((3, 4), 'something', of_type=(int, int, int))

    with pytest.raises(CheckError):
        check.tuple_param((3, 4), 'something', of_type=(int, str))

    with pytest.raises(CheckError):
        check.tuple_param((3, 4), 'something', of_type=(str, str))


def test_matrix_param():
    assert check.matrix_param([[1, 2], [2, 3]], 'something')

    with pytest.raises(CheckError):
        assert check.matrix_param(None, 'something')

    with pytest.raises(CheckError):
        assert check.matrix_param([1, 2, 4], 'something')

    with pytest.raises(CheckError):
        assert check.matrix_param([], 'something')

    with pytest.raises(CheckError):
        assert check.matrix_param([[1, 2], 3], 'soemthing')

    with pytest.raises(CheckError):
        assert check.matrix_param([[1, 2], [3.0, 4.1]], 'something', of_type=int)

    with pytest.raises(CheckError):
        assert check.matrix_param([[1, 2], [2, 3, 4]], 'something')


def test_opt_tuple_param():
    assert check.opt_tuple_param((1, 2), 'something')
    assert check.opt_tuple_param(None, 'something') is None
    assert check.opt_tuple_param(None, 'something', (2)) == (2)

    with pytest.raises(CheckError):
        check.opt_tuple_param(1, 'something')

    with pytest.raises(CheckError):
        check.opt_tuple_param([1], 'something')

    with pytest.raises(CheckError):
        check.opt_tuple_param({1: 2}, 'something')

    with pytest.raises(CheckError):
        check.opt_tuple_param('kdjfkd', 'something')

    assert check.opt_tuple_param((3, 4), 'something', of_type=int)
    assert check.opt_tuple_param(('foo', 'bar'), 'something', of_type=str)

    assert check.opt_tuple_param((3, 4), 'something', of_type=(int, int))
    assert check.opt_tuple_param((3, 4), 'something', of_type=(int, int))
    assert check.opt_tuple_param((3, 'bar'), 'something', of_type=(int, str))

    with pytest.raises(CheckError):
        check.opt_tuple_param((3, 4, 5), 'something', of_type=(int, int))

    with pytest.raises(CheckError):
        check.opt_tuple_param((3, 4), 'something', of_type=(int, int, int))

    with pytest.raises(CheckError):
        check.opt_tuple_param((3, 4), 'something', of_type=(int, str))

    with pytest.raises(CheckError):
        check.opt_tuple_param((3, 4), 'something', of_type=(str, str))


def test_opt_type_param():
    class Foo(object):
        pass

    assert check.opt_type_param(int, 'foo')
    assert check.opt_type_param(Foo, 'foo')

    assert check.opt_type_param(None, 'foo') is None
    assert check.opt_type_param(None, 'foo', Foo) is Foo

    with pytest.raises(CheckError):
        check.opt_type_param(check, 'foo')

    with pytest.raises(CheckError):
        check.opt_type_param(234, 'foo')

    with pytest.raises(CheckError):
        check.opt_type_param('bar', 'foo')

    with pytest.raises(CheckError):
        check.opt_type_param(Foo(), 'foo')


def test_type_param():
    class Bar(object):
        pass

    assert check.type_param(int, 'foo')
    assert check.type_param(Bar, 'foo')

    with pytest.raises(CheckError):
        check.type_param(None, 'foo')

    with pytest.raises(CheckError):
        check.type_param(check, 'foo')

    with pytest.raises(CheckError):
        check.type_param(234, 'foo')

    with pytest.raises(CheckError):
        check.type_param('bar', 'foo')

    with pytest.raises(CheckError):
        check.type_param(Bar(), 'foo')


def test_subclass_param():
    class Super(object):
        pass

    class Sub(Super):
        pass

    class Alone(object):
        pass

    assert check.subclass_param(Sub, 'foo', Super)

    with pytest.raises(CheckError):
        assert check.subclass_param(Alone, 'foo', Super)

    with pytest.raises(CheckError):
        assert check.subclass_param('value', 'foo', Super)

    assert check.opt_subclass_param(Sub, 'foo', Super)
    assert check.opt_subclass_param(None, 'foo', Super) is None

    with pytest.raises(CheckError):
        assert check.opt_subclass_param(Alone, 'foo', Super)

    with pytest.raises(CheckError):
        assert check.opt_subclass_param('value', 'foo', Super)


@contextmanager
def raises_with_message(exc_type, message_text):
    with pytest.raises(exc_type) as exc_info:
        yield

    assert str(exc_info.value) == message_text


def is_python_three():
    return sys.version_info[0] >= 3


def test_two_dim_dict():
    assert check.two_dim_dict_param({}, 'foo') == {}
    assert check.two_dim_dict_param({'key': {}}, 'foo')
    assert check.two_dim_dict_param({'key': {'key2': 2}}, 'foo')

    # make sure default dict passes
    default_dict = defaultdict(dict)
    default_dict['key']['key2'] = 2
    assert check.two_dim_dict_param(default_dict, 'foo')

    with raises_with_message(
        CheckError,
        '''Param "foo" is not a dict. Got None which is type <class 'NoneType'>.'''
        if is_python_three()
        else '''Param "foo" is not a dict. Got None which is type <type 'NoneType'>.''',
    ):
        check.two_dim_dict_param(None, 'foo')

    with raises_with_message(
        CheckError,
        "Value in dictionary mismatches expected type for key int_value. Expected value "
        "of type <class 'dict'>. Got value 2 of type <class 'int'>."
        if is_python_three()
        else "Value in dictionary mismatches expected type for key int_value. Expected value "
        "of type <type 'dict'>. Got value 2 of type <type 'int'>.",
    ):
        check.two_dim_dict_param({'int_value': 2}, 'foo')

    with raises_with_message(
        CheckError,
        "Value in dictionary mismatches expected type for key level_two_value_mismatch. "
        "Expected value of type (<class 'str'>,). Got value 2 of type <class 'int'>."
        if is_python_three()
        else "Value in dictionary mismatches expected type for key level_two_value_mismatch. "
        "Expected value of type (<type 'basestring'>,). Got value 2 of type <type 'int'>.",
    ):
        check.two_dim_dict_param(
            {'level_one_key': {'level_two_value_mismatch': 2}}, 'foo', value_type=str
        )

    with raises_with_message(
        CheckError,
        "Key in dictionary mismatches type. Expected <class 'int'>. Got 'key'"
        if is_python_three()
        else "Key in dictionary mismatches type. Expected <type 'int'>. Got 'key'",
    ):
        assert check.two_dim_dict_param({'key': {}}, 'foo', key_type=int)

    with raises_with_message(
        CheckError,
        "Key in dictionary mismatches type. Expected <class 'int'>. Got 'level_two_key'"
        if is_python_three()
        else "Key in dictionary mismatches type. Expected <type 'int'>. Got 'level_two_key'",
    ):
        assert check.two_dim_dict_param({1: {'level_two_key': 'something'}}, 'foo', key_type=int)


def test_opt_two_dim_dict_parm():
    assert check.opt_two_dim_dict_param({}, 'foo') == {}
    assert check.opt_two_dim_dict_param({'key': {}}, 'foo')
    assert check.opt_two_dim_dict_param({'key': {'key2': 2}}, 'foo')
    assert check.opt_two_dim_dict_param(None, 'foo') == {}

    with pytest.raises(CheckError):
        assert check.opt_two_dim_dict_param('str', 'foo')


def test_generator_param():
    def _test_gen():
        yield 1

    assert check.generator_param(_test_gen(), 'gen')

    gen = _test_gen()
    assert check.generator(gen)
    assert list(gen) == [1]
    assert check.generator(gen)
    assert list(gen) == []

    with pytest.raises(ParameterCheckError):
        assert check.generator_param(list(gen), 'gen')

    with pytest.raises(ParameterCheckError):
        assert check.generator_param(None, 'gen')

    with pytest.raises(ParameterCheckError):
        assert check.generator_param(_test_gen, 'gen')


def test_opt_generator_param():
    def _test_gen():
        yield 1

    assert check.opt_generator_param(_test_gen(), 'gen')

    assert check.opt_generator_param(None, 'gen') is None

    with pytest.raises(ParameterCheckError):
        assert check.opt_generator_param(_test_gen, 'gen')


def test_generator():
    def _test_gen():
        yield 1

    assert check.generator(_test_gen())

    gen = _test_gen()
    assert check.generator(gen)

    with pytest.raises(ParameterCheckError):
        assert check.generator(list(gen))

    with pytest.raises(ParameterCheckError):
        assert check.generator(None)

    with pytest.raises(ParameterCheckError):
        assert check.generator(_test_gen)


def test_opt_generator():
    def _test_gen():
        yield 1

    assert check.opt_generator(_test_gen())

    gen = _test_gen()
    assert check.opt_generator(gen)
    assert check.opt_generator(None) is None

    with pytest.raises(ParameterCheckError):
        assert check.opt_generator(list(gen))

    with pytest.raises(ParameterCheckError):
        assert check.opt_generator(_test_gen)


def test_internals():
    with pytest.raises(CheckError):
        check._check_key_value_types(None, str, str)  # pylint: disable=protected-access



import os

from dagster import RepositoryDefinition, pipeline, solid
from dagster.core.definitions.container import get_active_repository_data_from_image
from dagster.core.snap import active_repository_data_from_def
from dagster.serdes import serialize_dagster_namedtuple


def noop_repo():
    @solid
    def noop_solid(_):
        return 1

    @pipeline
    def noop_pipeline():
        noop_solid()

    return RepositoryDefinition(name='noop_repo', pipeline_defs=[noop_pipeline])


# pylint: disable=unused-argument
def mock_snapshot_provider(image, command, volumes):
    active_repo_data = active_repository_data_from_def(noop_repo())
    with open(
        os.path.abspath(os.path.join(list(volumes.keys())[0], 'asuperuniqueid.json')), 'w+'
    ) as fp:
        fp.write(serialize_dagster_namedtuple(active_repo_data))


def test_container_snapshot_provider(mocker):
    mocker.patch('dagster.core.definitions.container.uuid4', return_value='asuperuniqueid')

    execute_container_mock = mocker.patch(
        'dagster.core.definitions.container.run_serialized_container_command',
        side_effect=mock_snapshot_provider,
    )
    active_repository_data = get_active_repository_data_from_image("foo:latest")
    execute_container_mock.assert_called_with(
        image="foo:latest",
        command='dagster repository snapshot {}'.format(
            os.path.join('/data', 'asuperuniqueid.json')
        ),
        volumes=mocker.ANY,
    )
    assert active_repository_data == active_repository_data_from_def(noop_repo())



from collections import namedtuple

from dagster import check
from dagster.core.execution.plan.objects import StepInput, StepKind, StepOutput, StepOutputHandle
from dagster.core.execution.plan.plan import ExecutionPlan, ExecutionStep
from dagster.serdes import whitelist_for_serdes

from .pipeline_snapshot import PipelineIndex
from .utils import create_snapshot_id


def create_execution_plan_snapshot_id(execution_plan_snapshot):
    check.inst_param(execution_plan_snapshot, 'execution_plan_snapshot', ExecutionPlanSnapshot)
    return create_snapshot_id(execution_plan_snapshot)


class ExecutionPlanIndex:
    def __init__(self, execution_plan_snapshot, pipeline_index):
        self.execution_plan_snapshot = check.inst_param(
            execution_plan_snapshot, 'execution_plan_snapshot', ExecutionPlanSnapshot
        )
        self.pipeline_index = check.inst_param(pipeline_index, 'pipeline_index', PipelineIndex)

        self._step_index = {step.key: step for step in self.execution_plan_snapshot.steps}

        # https://github.com/dagster-io/dagster/issues/2442
        # check.invariant(
        #     execution_plan_snapshot.pipeline_snapshot_id == pipeline_index.pipeline_snapshot_id
        # )

    def has_step(self, key):
        check.str_param(key, 'key')
        return key in self._step_index

    def get_step_by_key(self, key):
        check.str_param(key, 'key')
        return self._step_index[key]

    @staticmethod
    def from_plan_and_index(execution_plan, pipeline_index):
        check.inst_param(execution_plan, 'execution_plan', ExecutionPlan)
        check.inst_param(pipeline_index, 'pipeline_index', PipelineIndex)
        return ExecutionPlanIndex(
            snapshot_from_execution_plan(
                execution_plan=execution_plan,
                pipeline_snapshot_id=pipeline_index.pipeline_snapshot_id,
            ),
            pipeline_index,
        )


@whitelist_for_serdes
class ExecutionPlanSnapshot(
    namedtuple('_ExecutionPlanSnapshot', 'steps artifacts_persisted pipeline_snapshot_id')
):
    def __new__(cls, steps, artifacts_persisted, pipeline_snapshot_id):
        return super(ExecutionPlanSnapshot, cls).__new__(
            cls,
            steps=check.list_param(steps, 'steps', of_type=ExecutionStepSnap),
            artifacts_persisted=check.bool_param(artifacts_persisted, 'artifacts_persisted'),
            pipeline_snapshot_id=check.str_param(pipeline_snapshot_id, 'pipeline_snapshot_id'),
        )


@whitelist_for_serdes
class ExecutionStepSnap(
    namedtuple('_ExecutionStepSnap', 'key inputs outputs solid_handle_id kind metadata_items')
):
    def __new__(cls, key, inputs, outputs, solid_handle_id, kind, metadata_items):
        return super(ExecutionStepSnap, cls).__new__(
            cls,
            key=check.str_param(key, 'key'),
            inputs=check.list_param(inputs, 'inputs', ExecutionStepInputSnap),
            outputs=check.list_param(outputs, 'outputs', ExecutionStepOutputSnap),
            solid_handle_id=check.str_param(solid_handle_id, 'solid_handle_id'),
            kind=check.inst_param(kind, 'kind', StepKind),
            metadata_items=check.list_param(
                metadata_items, 'metadata_items', ExecutionPlanMetadataItemSnap
            ),
        )


@whitelist_for_serdes
class ExecutionStepInputSnap(
    namedtuple('_ExecutionStepInputSnap', 'name dagster_type_key upstream_output_handles')
):
    def __new__(cls, name, dagster_type_key, upstream_output_handles):
        return super(ExecutionStepInputSnap, cls).__new__(
            cls,
            check.str_param(name, 'name'),
            check.str_param(dagster_type_key, 'dagster_type_key'),
            check.list_param(
                upstream_output_handles, 'upstream_output_handles', of_type=StepOutputHandle
            ),
        )

    @property
    def upstream_step_keys(self):
        return [output_handle.step_key for output_handle in self.upstream_output_handles]


@whitelist_for_serdes
class ExecutionStepOutputSnap(namedtuple('_ExecutionStepOutputSnap', 'name dagster_type_key')):
    def __new__(cls, name, dagster_type_key):
        return super(ExecutionStepOutputSnap, cls).__new__(
            cls,
            check.str_param(name, 'name'),
            check.str_param(dagster_type_key, 'dagster_type_key'),
        )


@whitelist_for_serdes
class ExecutionPlanMetadataItemSnap(namedtuple('_ExecutionPlanMetadataItemSnap', 'key value')):
    def __new__(cls, key, value):
        return super(ExecutionPlanMetadataItemSnap, cls).__new__(
            cls, check.str_param(key, 'key'), check.str_param(value, 'value'),
        )


def _snapshot_from_step_input(step_input):
    check.inst_param(step_input, 'step_input', StepInput)
    return ExecutionStepInputSnap(
        name=step_input.name,
        dagster_type_key=step_input.dagster_type.key,
        upstream_output_handles=step_input.source_handles,
    )


def _snapshot_from_step_output(step_output):
    check.inst_param(step_output, 'step_output', StepOutput)
    return ExecutionStepOutputSnap(
        name=step_output.name, dagster_type_key=step_output.dagster_type.key
    )


def _snapshot_from_execution_step(execution_step):
    check.inst_param(execution_step, 'execution_step', ExecutionStep)
    return ExecutionStepSnap(
        key=execution_step.key,
        inputs=sorted(
            list(map(_snapshot_from_step_input, execution_step.step_inputs)), key=lambda si: si.name
        ),
        outputs=sorted(
            list(map(_snapshot_from_step_output, execution_step.step_outputs)),
            key=lambda so: so.name,
        ),
        solid_handle_id=execution_step.solid_handle.to_string(),
        kind=execution_step.kind,
        metadata_items=sorted(
            [
                ExecutionPlanMetadataItemSnap(key=key, value=value)
                for key, value in execution_step.tags.items()
            ],
            key=lambda md: md.key,
        )
        if execution_step.tags
        else [],
    )


def snapshot_from_execution_plan(execution_plan, pipeline_snapshot_id):
    check.inst_param(execution_plan, 'execution_plan', ExecutionPlan)
    check.str_param(pipeline_snapshot_id, 'pipeline_snapshot_id')

    return ExecutionPlanSnapshot(
        steps=sorted(
            list(map(_snapshot_from_execution_step, execution_plan.steps)), key=lambda es: es.key
        ),
        artifacts_persisted=execution_plan.artifacts_persisted,
        pipeline_snapshot_id=pipeline_snapshot_id,
    )



from functools import update_wrapper

from dagster import check
from dagster.builtins import Int
from dagster.config.field import Field
from dagster.config.field_utils import check_user_facing_opt_config_param
from dagster.core.errors import DagsterUnmetExecutorRequirementsError
from dagster.core.execution.config import InProcessExecutorConfig, MultiprocessExecutorConfig
from dagster.core.execution.retries import Retries, get_retries_config


class ExecutorDefinition(object):
    '''
    Args:
        name (Optional[str]): The name of the executor.
        config (Optional[Any]): The schema for the config. Configuration data available in
            `init_context.executor_config`.

            This value can be any of:

            1. A Python primitive type that resolves to a Dagster config type
               (:py:class:`~python:int`, :py:class:`~python:float`, :py:class:`~python:bool`,
               :py:class:`~python:str`, or :py:class:`~python:list`).

            2. A Dagster config type: :py:data:`~dagster.Int`, :py:data:`~dagster.Float`,
               :py:data:`~dagster.Bool`, :py:data:`~dagster.String`,
               :py:data:`~dagster.StringSource`, :py:data:`~dagster.Path`, :py:data:`~dagster.Any`,
               :py:class:`~dagster.Array`, :py:data:`~dagster.Noneable`, :py:data:`~dagster.Enum`,
               :py:class:`~dagster.Selector`, :py:class:`~dagster.Shape`, or
               :py:class:`~dagster.Permissive`.

            3. A bare python dictionary, which will be automatically wrapped in
               :py:class:`~dagster.Shape`. Values of the dictionary are resolved recursively
               according to the same rules.

            4. A bare python list of length one which itself is config type.
               Becomes :py:class:`Array` with list element as an argument.

            5. An instance of :py:class:`~dagster.Field`.

        executor_creation_fn(Optional[Callable]): Should accept an :py:class:`InitExecutorContext`
            and return an instance of :py:class:`ExecutorConfig`.
        required_resource_keys (Optional[Set[str]]): Keys for the resources required by the
            executor.

    '''

    def __init__(self, name, config=None, executor_creation_fn=None, required_resource_keys=None):
        self._name = check.str_param(name, 'name')
        self._config_field = check_user_facing_opt_config_param(config, 'config')
        self._executor_creation_fn = check.opt_callable_param(
            executor_creation_fn, 'executor_creation_fn'
        )
        self._required_resource_keys = frozenset(
            check.opt_set_param(required_resource_keys, 'required_resource_keys', of_type=str)
        )

    @property
    def name(self):
        return self._name

    @property
    def config_field(self):
        return self._config_field

    @property
    def executor_creation_fn(self):
        return self._executor_creation_fn

    @property
    def required_resource_keys(self):
        return self._required_resource_keys


def executor(name=None, config=None, required_resource_keys=None):
    '''Define an executor.

    The decorated function should accept an :py:class:`InitExecutorContext` and return an instance
    of :py:class:`ExecutorConfig`.

    Args:
        name (Optional[str]): The name of the executor.
        config (Optional[Any]): The schema for the config. Configuration data available in
            `init_context.executor_config`.

            This value can be any of:

            1. A Python primitive type that resolves to a Dagster config type
               (:py:class:`~python:int`, :py:class:`~python:float`, :py:class:`~python:bool`,
               :py:class:`~python:str`, or :py:class:`~python:list`).

            2. A Dagster config type: :py:data:`~dagster.Int`, :py:data:`~dagster.Float`,
               :py:data:`~dagster.Bool`, :py:data:`~dagster.String`,
               :py:data:`~dagster.StringSource`, :py:data:`~dagster.Path`, :py:data:`~dagster.Any`,
               :py:class:`~dagster.Array`, :py:data:`~dagster.Noneable`, :py:data:`~dagster.Enum`,
               :py:class:`~dagster.Selector`, :py:class:`~dagster.Shape`, or
               :py:class:`~dagster.Permissive`.

            3. A bare python dictionary, which will be automatically wrapped in
               :py:class:`~dagster.Shape`. Values of the dictionary are resolved recursively
               according to the same rules.

            4. A bare python list of length one which itself is config type.
               Becomes :py:class:`Array` with list element as an argument.

            5. An instance of :py:class:`~dagster.Field`.

        required_resource_keys (Optional[Set[str]]): Keys for the resources required by the
            executor.
    '''
    if callable(name):
        check.invariant(config is None)
        check.invariant(required_resource_keys is None)
        return _ExecutorDecoratorCallable()(name)

    return _ExecutorDecoratorCallable(
        name=name, config=config, required_resource_keys=required_resource_keys,
    )


class _ExecutorDecoratorCallable(object):
    def __init__(self, name=None, config=None, required_resource_keys=None):
        self.name = check.opt_str_param(name, 'name')
        self.config = config  # type check in definition
        self.required_resource_keys = required_resource_keys  # type check in definition

    def __call__(self, fn):
        check.callable_param(fn, 'fn')

        if not self.name:
            self.name = fn.__name__

        executor_def = ExecutorDefinition(
            name=self.name,
            config=self.config,
            executor_creation_fn=fn,
            required_resource_keys=self.required_resource_keys,
        )

        update_wrapper(executor_def, wrapped=fn)

        return executor_def


@executor(
    name='in_process',
    config={'retries': get_retries_config(), 'marker_to_close': Field(str, is_required=False),},
)
def in_process_executor(init_context):
    '''The default in-process executor.

    In most Dagster environments, this will be the default executor. It is available by default on
    any :py:class:`ModeDefinition` that does not provide custom executors. To select it explicitly,
    include the following top-level fragment in config:

    .. code-block:: yaml

        execution:
          in_process:

    Execution priority can be configured using the ``dagster/priority`` tag via solid metadata,
    where the higher the number the higher the priority. 0 is the default and both positive
    and negative numbers can be used.
    '''
    from dagster.core.engine.init import InitExecutorContext

    check.inst_param(init_context, 'init_context', InitExecutorContext)

    return InProcessExecutorConfig(
        # shouldn't need to .get() here - issue with defaults in config setup
        retries=Retries.from_config(init_context.executor_config.get('retries', {'enabled': {}})),
        marker_to_close=init_context.executor_config.get('marker_to_close'),
    )


@executor(
    name='multiprocess',
    config={
        'max_concurrent': Field(Int, is_required=False, default_value=0),
        'retries': get_retries_config(),
    },
)
def multiprocess_executor(init_context):
    '''The default multiprocess executor.

    This simple multiprocess executor is available by default on any :py:class:`ModeDefinition`
    that does not provide custom executors. To select the multiprocess executor, include a fragment
    such as the following in your config:

    .. code-block:: yaml

        execution:
          multiprocess:
            max_concurrent: 4

    The ``max_concurrent`` arg is optional and tells the execution engine how many processes may run
    concurrently. By default, or if you set ``max_concurrent`` to be 0, this is the return value of
    :py:func:`python:multiprocessing.cpu_count`.

    Execution priority can be configured using the ``dagster/priority`` tag via solid metadata,
    where the higher the number the higher the priority. 0 is the default and both positive
    and negative numbers can be used.
    '''
    from dagster.core.definitions.handle import ExecutionTargetHandle
    from dagster.core.engine.init import InitExecutorContext

    check.inst_param(init_context, 'init_context', InitExecutorContext)

    check_cross_process_constraints(init_context)

    # ExecutionTargetHandle.get_handle returns an ExecutionTargetHandleCacheEntry, which is a tuple
    # (handle, solid_subset). Right now we are throwing away the solid_subset that we store in the
    # cache -- this is fragile and we should fix this with
    # https://github.com/dagster-io/dagster/issues/2115 and friends so there are not multiple
    # sources of truth for the solid subset
    handle, _ = ExecutionTargetHandle.get_handle(init_context.pipeline_def)
    return MultiprocessExecutorConfig(
        handle=handle,
        max_concurrent=init_context.executor_config['max_concurrent'],
        retries=Retries.from_config(init_context.executor_config['retries']),
    )


default_executors = [in_process_executor, multiprocess_executor]


def check_cross_process_constraints(init_context):
    from dagster.core.engine.init import InitExecutorContext

    check.inst_param(init_context, 'init_context', InitExecutorContext)

    _check_pipeline_has_target_handle(init_context.pipeline_def)
    _check_non_ephemeral_instance(init_context.instance)
    _check_persistent_storage_requirement(init_context.system_storage_def)


def _check_pipeline_has_target_handle(pipeline_def):
    from dagster.core.definitions.handle import ExecutionTargetHandle

    handle, _ = ExecutionTargetHandle.get_handle(pipeline_def)
    if not handle:
        raise DagsterUnmetExecutorRequirementsError(
            'You have attempted to use an executor that uses multiple processes with the pipeline "{name}" '
            'that can not be re-hydrated. Pipelines must be loaded in a way that allows dagster to reconstruct '
            'them in a new process. This means: \n'
            '  * using the file, module, or repository.yaml arguments of dagit/dagster-graphql/dagster\n'
            '  * constructing an ExecutionTargetHandle directly\n'.format(name=pipeline_def.name)
        )


def _check_persistent_storage_requirement(system_storage_def):
    if not system_storage_def.is_persistent:
        raise DagsterUnmetExecutorRequirementsError(
            (
                'You have attempted to use an executor that uses multiple processes while using system '
                'storage {storage_name} which does not persist intermediates. '
                'This means there would be no way to move data between different '
                'processes. Please configure your pipeline in the storage config '
                'section to use persistent system storage such as the filesystem.'
            ).format(storage_name=system_storage_def.name)
        )


def _check_non_ephemeral_instance(instance):
    if instance.is_ephemeral:
        raise DagsterUnmetExecutorRequirementsError(
            'You have attempted to use an executor that uses multiple processes with an '
            'ephemeral DagsterInstance. A non-ephemeral instance is needed to coordinate '
            'execution between multiple processes. You can configure your default instance '
            'via $DAGSTER_HOME or ensure a valid one is passed when invoking the python APIs.'
        )



import hashlib

from dagster.serdes import serialize_dagster_namedtuple


def create_snapshot_id(snapshot):
    json_rep = serialize_dagster_namedtuple(snapshot)
    m = hashlib.sha1()  # so that hexdigest is 40, not 64 bytes
    m.update(json_rep.encode())
    return m.hexdigest()



from dagster.core.instance import DagsterInstance, InstanceRef
from dagster.core.snap import (
    ExecutionPlanIndex,
    PipelineIndex,
    create_execution_plan_snapshot_id,
    create_pipeline_snapshot_id,
)
from dagster.utils import file_relative_path
from dagster.utils.test import restore_directory


# a change of schema in the snapshot hierarchy caused hashes to be different
# when snapshots reloaded
def test_0_7_10_snapshot_id_change():
    test_dir = file_relative_path(__file__, 'snapshot_0_7_9_shapshot_id_creation_change/sqlite')
    with restore_directory(test_dir):

        instance = DagsterInstance.from_ref(InstanceRef.from_dir(test_dir))
        # run_id = 'e297fa70-49e8-43f8-abfe-1634f02644f6'

        old_pipeline_snapshot_id = '88528edde2ed64da3c39cca0da8ba2f7586c1a5d'
        old_execution_plan_snapshot_id = '2246f8e5a10d21e15fbfa3773d7b2d0bc1fa9d3d'

        pipeline_snapshot = instance.get_pipeline_snapshot(old_pipeline_snapshot_id)
        ep_snapshot = instance.get_execution_plan_snapshot(old_execution_plan_snapshot_id)

        # It is the pipeline snapshot that changed
        # Verify that snapshot ids are not equal
        assert create_pipeline_snapshot_id(pipeline_snapshot) != old_pipeline_snapshot_id

        assert create_execution_plan_snapshot_id(ep_snapshot) == old_execution_plan_snapshot_id

        # This previously failed with a check error
        assert ExecutionPlanIndex(ep_snapshot, PipelineIndex(pipeline_snapshot))


# Scripts to create this (run against 0.7.9)
#
# from dagster import pipeline, solid, DagsterInstance, execute_pipeline
# from dagster.core.snap.utils import create_snapshot_id
#
# from dagster.serdes import serialize_pp
#
# @solid
# def noop_solid(_):
#     pass
#
# @pipeline
# def noop_pipeline():
#     noop_solid()
#
# instance = DagsterInstance.get()
#
# result = execute_pipeline(noop_pipeline, instance=instance)
#
# run_id = result.run_id

# print(serialize_pp(run))
